/******************************************************************************
 *
 * Source file for harness-test (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

#define HARNESS_TEST_PROCESSES_BUFFER 1024
#define HARNESS_TEST_PROCESSES_COUNT  64
#define HARNESS_TEST_PLUGINS_COUNT    64

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness-test.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Harness-test plug-in module data type. */
typedef struct {
  pthread_mutex_t  mutex;     /* mutex                 */
  unsigned int     count;     /* instances array count */
  struct {
    unsigned int   handle;    /* instance handle       */
  }               *instances; /* instances array       */
} harness_test_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the harness-test plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_INIT(harness_test_init);

/*
 * Finalizes the harness-test plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_INIT(harness_test_fini);

/*
 * Tests the process launcher functions of the kernel.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_test_processes ();

/*
 * Tests the plug-in loader functions of the kernel.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_test_plugins ();

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Harness-test plug-in module data. */
harness_test_data_t harness_test_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex                 */
  0,                         /* instances array count */
  NULL                       /* instances array       */
};

/* Harness-test plug-in "object". */
harness_test_t harness_test = {
  {
    LIBHARNESS_TEST_VERSION_CURRENT,  /* current version         */
    LIBHARNESS_TEST_VERSION_REVISION, /* current revision        */
    LIBHARNESS_TEST_VERSION_AGE,      /* version age             */
    LIBHARNESS_TEST_VERSION_FIRST     /* first supported version */
  },                                  /* plug-in version         */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the harness-test plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_init"
HARNESS_PLUGINS_INIT(harness_test_init) {
  int           error;
  void         *instances;
  unsigned int  index;
  /* Lock harness-test plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_test_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to lock harness-test plug-in mutex)))
    harness_syserr();
    return -1;
  }
  /* Search for handle in instances array. */
  for (index = 0; index < harness_test_data.count; index++) {
    if (harness_test_data.instances[index].handle == handle) {
      HARNESS_LOG((
        HARNESS_WARN(handle is already in instances array)))
      harness_syserr();
      /* Unlock harness-test plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_test_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-test plug-in mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Increase instances array. */
  index = harness_test_data.count;
  harness_test_data.count++;
  /* Reallocate instances array. */
  if (NULL == (instances = realloc(harness_test_data.instances,
                                   harness_test_data.count *
                            sizeof(harness_test_data.instances[0])))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to reallocate instances array)))
    /* Unlock harness-test plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_test_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-test plug-in mutex)))
      harness_syserr();
    }
    return -1;
  }
  harness_test_data.instances = instances;
  /* Save instances array entry. */
  harness_test_data.instances[index].handle = handle;
  /* Check for first initialization. */
  if (1 == harness_test_data.count) {
    /* Submit job to test kernel and forker. */
    if (0 != harness_threads_submit(&harness_test_run, NULL, NULL)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to submit job to test kernel and forker)))
      /* Reduce instances array. */
      harness_test_data.count = 0;
      /* Deallocate instances array. */
      free(harness_test_data.instances);
      harness_test_data.instances = NULL;
      /* Unlock harness-test plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_test_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-test plug-in mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Unlock harness-test plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_test_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-test plug-in mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Finalizes the harness-test plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_fini"
HARNESS_PLUGINS_INIT(harness_test_fini) {
  int           error;
  void         *instances;
  unsigned int  index;
  /* Lock harness-test plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_test_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to lock harness-test plug-in mutex)))
    harness_syserr();
    return -1;
  }
  /* Search for handle in instances array. */
  for (index = 0; index < harness_test_data.count; index++) {
    if (harness_test_data.instances[index].handle == handle) {
      break;
    }
  }
  /* Check if handle is not in instances array. */
  if (index == harness_test_data.count) {
    HARNESS_LOG((
      HARNESS_WARN(handle is not in instances array)))
    harness_syserr();
    /* Unlock harness-test plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_test_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-test plug-in mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Remove instance from instances array. */
  harness_test_data.count--;
  memmove(harness_test_data.instances + index,
          harness_test_data.instances + index + 1,
         (harness_test_data.count     - index) *
          sizeof(unsigned int));
  /* Reallocate instances array. */
  if (0 == harness_test_data.count) {
    free(harness_test_data.instances);
    harness_test_data.instances = NULL;
  } else if (NULL == (instances =
                      realloc(harness_test_data.instances,
                              harness_test_data.count *
                       sizeof(harness_test_data.instances[0])))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to reallocate instances array)))
  } else {
    harness_test_data.instances = instances;
  }
  /* Unlock harness-test plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_test_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-test plug-in mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Tests several functions of the kernel and subsequently the forker.
 *
 * argument = NULL
 * return   = 0 on success or -1 on any system error with errno set
 *            appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_run"
HARNESS_THREADS_JOB(harness_test_run) {
  HARNESS_LOG((HARNESS_INFO(harness-test startup)))
  /* Test process launcher. */
  if (0 != harness_test_processes()) {
    fprintf(stdout, "ERROR: Test of process launcher failed.\n");
    return -1;
  }
  /* Test plug-in loader. */
  if (0 != harness_test_plugins()) {
    fprintf(stdout, "ERROR: Test of plug-in loader failed.\n");
    return -1;
  }
  HARNESS_LOG((HARNESS_INFO(harness-test shutdown)))
  /* Shut down kernel. */
  harness_kernel_shutdown();
  return 0;
}

/*
 * Tests the process launcher functions of the kernel.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_processes"
int harness_test_processes () {
  int          status, signum;
  char         buffer1[HARNESS_TEST_PROCESSES_BUFFER],
               buffer2[HARNESS_TEST_PROCESSES_BUFFER];
  unsigned int index, handles[HARNESS_TEST_PROCESSES_COUNT];
  for (index = 0; index < HARNESS_TEST_PROCESSES_COUNT; index++) {
    /* Run test program. */
#ifdef DEBUG
    if (0 != harness_processes_execute(&handles[index],
                                        "harness-test.debug", NULL)) {
#else /* DEBUG */
    if (0 != harness_processes_execute(&handles[index],
                                        "harness-test", NULL)) {
#endif /* DEBUG */
      HARNESS_LOG((
        HARNESS_WARN(unable to execute test program)))
      return -1;
    }
  }
  for (index = 0; index < HARNESS_TEST_PROCESSES_COUNT; index++) {
    status = 0;
    /* Wait for test program exit. */
    if (0 != harness_processes_wait(handles[index], &status)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to wait for test program exit)))
      return -1;
    }
    /* Check for successful exit. */
    if ((0 == WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
      HARNESS_LOG((
        HARNESS_WARN(test program exited with error)))
      return -1;
    }
  }
  for (index = 0; index < HARNESS_TEST_PROCESSES_COUNT; index++) {
    status = 0;
    /* Terminate test program. */
    if (0 != harness_processes_terminate(handles[index], &status)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to terminate test program)))
      return -1;
    }
    /* Check for successful exit. */
    if ((0 == WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
      HARNESS_LOG((
        HARNESS_WARN(test program exited with error)))
      return -1;
    }
  }
  for (index = 0; index < HARNESS_TEST_PROCESSES_COUNT; index++) {
    /* Run test program. */
#ifdef DEBUG
    if (0 != harness_processes_execute(&handles[index],
                                        "harness-test.debug --pipe", NULL)) {
#else /* DEBUG */
    if (0 != harness_processes_execute(&handles[index],
                                        "harness-test --pipe", NULL)) {
#endif /* DEBUG */
      HARNESS_LOG((
        HARNESS_WARN(unable to execute test program)))
      return -1;
    }
  }
  for (index = 0; index < HARNESS_TEST_PROCESSES_COUNT; index++) {
    status = 0;
    /* Terminate test program. */
    if (0 != harness_processes_terminate(handles[index], &status)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to terminate test program)))
      return -1;
    }
    /* Check for successful exit. */
    if ((0 == WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
      HARNESS_LOG((
        HARNESS_WARN(test program exited with error)))
      return -1;
    }
  }
  /* Run test program. */
#ifdef DEBUG
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test.debug --pipe", NULL)) {
#else /* DEBUG */
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test --pipe", NULL)) {
#endif /* DEBUG */
    HARNESS_LOG((
      HARNESS_WARN(unable to execute test program)))
    return -1;
  }
  for (index = 0; index < HARNESS_TEST_PROCESSES_COUNT; index++) {
    /* Initialize buffers. */
    memset(buffer1, random(), HARNESS_TEST_PROCESSES_BUFFER);
    memset(buffer2,        0, HARNESS_TEST_PROCESSES_BUFFER);
    /* Write to test program. */
    if (0 != harness_processes_write(handles[0],
                                      HARNESS_TEST_PROCESSES_BUFFER,
                                      buffer1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write to test program)))
      return -1;
    }
    /* Read from test program. */
    if (HARNESS_TEST_PROCESSES_BUFFER !=
        harness_processes_read(handles[0], HARNESS_TEST_PROCESSES_BUFFER,
                                buffer2, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read from test program)))
      return -1;
    }
    /* Check for data correctness. */
    if (0 != memcmp(buffer1, buffer2, HARNESS_TEST_PROCESSES_BUFFER)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read correct data from test program)))
      return -1;
    }
  }
  status = 0;
  /* Terminate test program. */
  if (0 != harness_processes_terminate(handles[0], &status)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to terminate test program)))
    return -1;
  }
  /* Check for successful exit. */
  if ((0 == WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
    HARNESS_LOG((
      HARNESS_WARN(test program exited with error)))
    return -1;
  }
  /* Run test program. */
#ifdef DEBUG
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test.debug --echo test tset",
                                      NULL)) {
#else /* DEBUG */
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test --echo test tset", NULL)) {
#endif /* DEBUG */
    HARNESS_LOG((
      HARNESS_WARN(unable to execute test program)))
    return -1;
  }
  /* Read from test program. */
  if (9 != harness_processes_read(handles[0], 9, buffer2, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read from test program)))
    return -1;
  }
  /* Check for data correctness. */
  if (0 != memcmp("test tset", buffer2, 9)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read correct data from test program)))
    return -1;
  }
  status = 0;
  /* Terminate test program. */
  if (0 != harness_processes_terminate(handles[0], &status)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to terminate test program)))
    return -1;
  }
  /* Check for successful exit. */
  if ((0 == WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
    HARNESS_LOG((
      HARNESS_WARN(test program exited with error)))
    return -1;
  }
  /* Run test program. */
#ifdef DEBUG
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test.debug --env HARNESS_TEST",
                                      "HARNESS_TEST='test tset'")) {
#else /* DEBUG */
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test --env HARNESS_TEST",
                                      "HARNESS_TEST='test tset'")) {
#endif /* DEBUG */
    HARNESS_LOG((
      HARNESS_WARN(unable to execute test program)))
    return -1;
  }
  /* Read from test program. */
  if (9 != harness_processes_read(handles[0], 9, buffer2, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read from test program)))
    return -1;
  }
  /* Check for data correctness. */
  if (0 != memcmp("test tset", buffer2, 9)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read correct data from test program)))
    return -1;
  }
  status = 0;
  /* Terminate test program. */
  if (0 != harness_processes_terminate(handles[0], &status)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to terminate test program)))
    return -1;
  }
  /* Check for successful exit. */
  if ((0 == WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
    HARNESS_LOG((
      HARNESS_WARN(test program exited with error)))
    return -1;
  }
  /* Run test program. */
#ifdef DEBUG
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test.debug --signal", NULL)) {
#else /* DEBUG */
  if (0 != harness_processes_execute(&handles[0],
                                      "harness-test --signal", NULL)) {
#endif /* DEBUG */
    HARNESS_LOG((
      HARNESS_WARN(unable to execute test program)))
    return -1;
  }
  /* Signal test program. */
  if (0 != harness_processes_signal(handles[0], SIGUSR1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to signal test program)))
    return -1;
  }
  /* Read from test program. */
  if (sizeof(int) != harness_processes_read(handles[0], sizeof(int),
                                             &signum, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read from test program)))
    return -1;
  }
  /* Check for data correctness. */
  if (SIGUSR1 != signum) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read correct data from test program)))
    return -1;
  }
  /* Terminate test program. */
  if (0 != harness_processes_terminate(handles[0], &status)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to terminate test program)))
    return -1;
  }
  /* Check for successful exit. */
  if ((0 == WIFEXITED(status))||(EXIT_SUCCESS != WEXITSTATUS(status))) {
    HARNESS_LOG((
      HARNESS_WARN(test program exited with error)))
    return -1;
  }
  return 0;
}

/*
 * Tests the plug-in loader functions of the kernel.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_test_plugins"
int harness_test_plugins () {
  void         *symbol1, *symbol2;
  unsigned int  index, handles[HARNESS_TEST_PLUGINS_COUNT];
  for (index = 0; index < HARNESS_TEST_PLUGINS_COUNT; index++) {
    /* Load test plug-in. */
    if (0 != harness_plugins_load(&handles[index], "libharness-test", 0)) {
      HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
      return -1;
    }
  }
  for (index = 0; index < HARNESS_TEST_PLUGINS_COUNT; index++) {
    /* Unload test plug-in. */
    if (0 != harness_plugins_unload(handles[index])) {
      HARNESS_LOG((
        HARNESS_WARN(unable to unload test plug-in)))
      return -1;
    }
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load(&handles[0], "libharness-test.so", 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load(&handles[0], HARNESS_STRING(\
             libharness-test.LIBHARNESS_TEST_VERSION_CURRENT), 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load(&handles[0], HARNESS_STRING(\
             libharness-test.LIBHARNESS_TEST_VERSION_CURRENT.so), 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load(&handles[0], HARNESS_STRING(\
             libharness-test.LIBHARNESS_TEST_VERSION), 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load(&handles[0], HARNESS_STRING(\
             libharness-test.LIBHARNESS_TEST_VERSION.so), 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load2(&handles[0], "libharness-test",
                                  LIBHARNESS_TEST_VERSION_CURRENT, 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load2(&handles[0], "libharness-test.so",
                                  LIBHARNESS_TEST_VERSION_CURRENT, 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load3(&handles[0], "libharness-test",
                                  LIBHARNESS_TEST_VERSION_FIRST,
                                  LIBHARNESS_TEST_VERSION_AGE,
                                  LIBHARNESS_TEST_VERSION_REVISION, 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load3(&handles[0], "libharness-test.so",
                                  LIBHARNESS_TEST_VERSION_FIRST,
                                  LIBHARNESS_TEST_VERSION_AGE,
                                  LIBHARNESS_TEST_VERSION_REVISION, 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load4(&handles[0], "libharness-test",
                                  LIBHARNESS_TEST_VERSION_CURRENT,
                                  LIBHARNESS_TEST_VERSION_AGE,
                                  LIBHARNESS_TEST_VERSION_REVISION, 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load4(&handles[0], "libharness-test.so",
                                  LIBHARNESS_TEST_VERSION_CURRENT,
                                  LIBHARNESS_TEST_VERSION_AGE,
                                  LIBHARNESS_TEST_VERSION_REVISION, 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load(&handles[0], "libharness-test", 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Load test plug-in. */
  if (0 != harness_plugins_load(&handles[1], "libharness-test", 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load test plug-in)))
    return -1;
  }
  /* Get test plug-in symbol. */
  if (0 != harness_plugins_symbol(&symbol1, handles[0], "harness_test")) {
    HARNESS_LOG((
      HARNESS_WARN(unable to get test plug-in symbol)))
    return -1;
  }
  /* Get test plug-in symbol. */
  if (0 != harness_plugins_symbol(&symbol2, handles[1], "harness_test")) {
    HARNESS_LOG((
      HARNESS_WARN(unable to get test plug-in symbol)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[1])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Unload test plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to unload test plug-in)))
    return -1;
  }
  /* Compare test plug-in symbols. */
  if (symbol1 != symbol2) {
    HARNESS_LOG((
      HARNESS_WARN(test plug-in symbols are not equal)))
    return -1;
  }
  /* Compare test plug-in symbols. */
  if (symbol1 != &harness_test) {
    HARNESS_LOG((
      HARNESS_WARN(test plug-in symbols do not match)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (0 != harness_plugins_load(&handles[0], "libharness-dummy",
                                 HARNESS_PLUGINS_UNIQUE)) {
    HARNESS_LOG((HARNESS_WARN(unable to load dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (-1 != harness_plugins_load(&handles[1], "libharness-dummy", 0)) {
    HARNESS_LOG((HARNESS_WARN(able to load dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (-1 != harness_plugins_load(&handles[1], "libharness-dummy",
                                  HARNESS_PLUGINS_UNIQUE)) {
    HARNESS_LOG((HARNESS_WARN(able to load dummy plug-in)))
    return -1;
  }
  /* Get dummy plug-in symbol. */
  if (0 != harness_plugins_symbol(&symbol1, handles[0], "harness_dummy")) {
    HARNESS_LOG((
      HARNESS_WARN(unable to get dummy plug-in symbol)))
    return -1;
  }
  /* Get global dummy plug-in symbol. */
  if (NULL != (symbol2 = dlsym(NULL, "harness_dummy"))) {
    HARNESS_LOG((
      HARNESS_WARN(able to get global dummy plug-in symbol)))
    return -1;
  }
  /* Unload dummy plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((
      HARNESS_WARN(unable to unload dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (0 != harness_plugins_load(&handles[0], "libharness-dummy", 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (-1 != harness_plugins_load(&handles[1], "libharness-dummy",
                                  HARNESS_PLUGINS_UNIQUE)) {
    HARNESS_LOG((HARNESS_WARN(able to load dummy plug-in)))
    return -1;
  }
  /* Unload dummy plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((
      HARNESS_WARN(unable to unload dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (0 != harness_plugins_load(&handles[0], "libharness-dummy",
                                 HARNESS_PLUGINS_EXPORT)) {
    HARNESS_LOG((HARNESS_WARN(unable to load dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (-1 != harness_plugins_load(&handles[1], "libharness-dummy", 0)) {
    HARNESS_LOG((HARNESS_WARN(able to load dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (-1 != harness_plugins_load(&handles[1], "libharness-dummy",
                                  HARNESS_PLUGINS_EXPORT)) {
    HARNESS_LOG((HARNESS_WARN(able to load dummy plug-in)))
    return -1;
  }
  /* Get dummy plug-in symbol. */
  if (0 != harness_plugins_symbol(&symbol1, handles[0], "harness_dummy")) {
    HARNESS_LOG((
      HARNESS_WARN(unable to get dummy plug-in symbol)))
    return -1;
  }
  /* Get global dummy plug-in symbol. */
  if (NULL == (symbol2 = dlsym(NULL, "harness_dummy"))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to get global dummy plug-in symbol)))
    return -1;
  }
  /* Compare test plug-in symbols. */
  if (symbol1 != symbol2) {
    HARNESS_LOG((
      HARNESS_WARN(dummy plug-in symbols are not equal)))
    return -1;
  }
  /* Unload dummy plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((
      HARNESS_WARN(unable to unload dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (0 != harness_plugins_load(&handles[0], "libharness-dummy", 0)) {
    HARNESS_LOG((HARNESS_WARN(unable to load dummy plug-in)))
    return -1;
  }
  /* Load dummy plug-in. */
  if (-1 != harness_plugins_load(&handles[1], "libharness-dummy",
                                  HARNESS_PLUGINS_EXPORT)) {
    HARNESS_LOG((HARNESS_WARN(able to load dummy plug-in)))
    return -1;
  }
  /* Unload dummy plug-in. */
  if (0 != harness_plugins_unload(handles[0])) {
    HARNESS_LOG((
      HARNESS_WARN(unable to unload dummy plug-in)))
    return -1;
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
