/******************************************************************************
 *
 * Source file for harness forker (forker process) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Forker process data type. */
typedef struct {
  unsigned int    wait;      /* wait flag           */
  unsigned int    exit;      /* exit flag           */
  int             pipes[2];  /* I/O pipe ends       */
  int             wakeup[2]; /* wake-up pipe        */
  unsigned int    envc;      /* default envc        */
  unsigned int    count;     /* process table count */
  struct {
    unsigned int  handle;    /* process handle      */
    pid_t         pid;       /* process system pid  */
    int           pipes[2];  /* process I/O pipes   */
  }              *procs;     /* process table array */
  unsigned int    size;      /* buffer size         */
  char           *buffer;    /* buffer              */
  sigset_t        sigmask;   /* signal mask         */
} harness_forker_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Sets the forker process configuration and runs the forker process.
 *
 * size   = The forker process buffer size.
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_main (unsigned int size,
                         int          pipes[2]);

/*
 * Runs the forker process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_main2 (int pipes[2]);

/*
 * Sets the forker process configuration and initializes the forker process.
 *
 * size   = The forker process buffer size.
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_init (unsigned int size,
                         int          pipes[2]);

/*
 * Initializes the forker process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_init2 (int pipes[2]);

/*
 * Finalizes the forker process.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_fini ();

/*
 * The signal handler of the forker process.
 *
 * signum = The number of the received signal.
 */
void harness_forker_handler (int signum);

/*
 * Runs the forker process tasks.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_run ();

/*
 * Processes kernel messages.
 *
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_forker_process ();

/*
 * Reads an execution request.
 *
 * handle = The process handle (return).
 * argc   = The argument vector count (return).
 * envc   = The environment vector count (return).
 * argv   = The argument vector (argument[0] is the program name) (return).
 * envv   = The environment vector (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_forker_read_execute (unsigned int   *handle,
                                 unsigned int   *argc,
                                 unsigned int   *envc,
                                 char         ***argv,
                                 char         ***envv);

/*
 * Processes an execution request.
 *
 * handle = The process handle.
 * argc   = The argument vector count.
 * envc   = The environment vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envv   = The environment vector.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_forker_process_execute (unsigned int   handle,
                                    unsigned int   argc,
                                    unsigned int   envc,
                                    char         **argv,
                                    char         **envv);

/*
 * Reads an input request.
 *
 * handle = The process handle (return).
 * length = The process input length (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_forker_read_input (unsigned int  *handle,
                               unsigned int  *length);

/*
 * Processes process input from the kernel.
 *
 * handle = The process handle (return).
 * length = The process input length (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_forker_process_input (unsigned int handle,
                                  unsigned int length);

/*
 * Forwards process output to the kernel.
 *
 * length = The index in the forker process table.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_forker_forward (unsigned int index);

/*
 * Waits for exited child processes..
 *
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
int harness_forker_wait ();

/*
 * Reads a buffer from a file descriptor (non-/blocking).
 *
 * fd     = The file descriptor to read from.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed file descriptor or -1
 *          on any other error with errno set appropriately.
 */
int harness_forker_read (int           fd,
                         unsigned int  length,
                         void         *buffer,
                         unsigned int  block);

/*
 * Writes a buffer into a file descriptor (blocking).
 *
 * fd     = The file descriptor to write to.
 * length = The buffer length to write.
 * buffer = The buffer to write from.
 * return = 0 on success, -2 on closed file descriptor or -1 on any other error
 *          with errno set appropriately.
 */
int harness_forker_write (int           fd,
                          unsigned int  length,
                          const void   *buffer);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Forker process data. */
harness_forker_data_t harness_forker_data = {
  0,                           /* wait flag           */
  0,                           /* exit flag           */
  { -1, -1 },                  /* I/O pipe ends       */
  { -1, -1 },                  /* wake-up pipe        */
  0,                           /* default envc        */
  0,                           /* process table count */
  NULL,                        /* process table array */
  HARNESS_FORKER_BUFFER_SIZE, /* buffer size         */
  NULL                         /* buffer              */
  /* not initialized */        /* signal mask         */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Sets the forker process buffer size.
 * This function may only be called before running harness_forker_main.
 *
 * size   = The forker process buffer size.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_size_set"
int harness_forker_size_set (unsigned int size) {
  /* Check size parameter. */
  if (0 == size) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(size parameter is 0)))
    return -1;
  }
  harness_forker_data.size = size;
  return 0;
}

/*
 * Increases/Decreases the forker process buffer size.
 * This function may only be called before initializing.
 *
 * size   = The forker process buffer size increment/decrement.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_size_add"
int harness_forker_size_add (int size) {
  size += harness_forker_data.size;
  /* Check new size. */
  if (0 == size) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(size is 0)))
    return -1;
  }
  if (0 > size) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(size is negative)))
    return -1;
  }
  return 0;
}

/*
 * Gets the forker process buffer size.
 *
 * size   = The forker process buffer size (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_size_get"
int harness_forker_size_get (unsigned int *size) {
#ifdef DEBUG
  /* Check size parameter. */
  if (NULL == size) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(size parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  *size = harness_forker_data.size;
  return 0;
}

/*
 * Sets the forker process configuration and runs the forker process.
 *
 * size   = The forker process buffer size.
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_main"
int harness_forker_main (unsigned int size,
                         int          pipes[2]) {
  /* Initialize forker process. */
  if (0 != harness_forker_init(size, pipes)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize forker process)))
    return -1;
  }
  /* Run forker process. */
  if (0 != harness_forker_run()) {
    HARNESS_LOG((HARNESS_WARN(unable to run forker process)))
    if (0 != harness_forker_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize forker process)))
    }
    return -1;
  }
  /* Finalize forker process. */
  if (0 != harness_forker_fini()) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize forker process)))
    return -1;
  }
  return 0;
}

/*
 * Runs the forker process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_main2"
int harness_forker_main2 (int pipes[2]) {
  /* Initialize forker process. */
  if (0 != harness_forker_init2(pipes)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize forker process)))
    return -1;
  }
  /* Run forker process. */
  if (0 != harness_forker_run()) {
    HARNESS_LOG((HARNESS_WARN(unable to run forker process)))
    if (0 != harness_forker_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize forker process)))
    }
    return -1;
  }
  /* Finalize forker process. */
  if (0 != harness_forker_fini()) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize forker process)))
    return -1;
  }
  return 0;
}

/*
 * Sets the forker process configuration and initializes the forker process.
 *
 * size   = The forker process buffer size.
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_init"
int harness_forker_init (unsigned int size,
                         int          pipes[2]) {
  /* Set the forker process buffer size. */
  if (0 != harness_threads_pool_minimum_set(size)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the forker process buffer size)))
    return -1;
  }
  /* Initialize the forker process. */
  if (0 != harness_forker_init2(pipes)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize the forker process)))
    return -1;
  }
  return 0;
}

/*
 * Initializes the forker process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_init2"
int harness_forker_init2 (int pipes[2]) {
  sigset_t sigmask;
  HARNESS_LOG((HARNESS_INFO(forker process startup)))
  /* Initialize I/O pipe ends. */
  harness_forker_data.pipes[0] = pipes[0];
  harness_forker_data.pipes[1] = pipes[1];
  /* Initialize wait and exit flags. */
  harness_forker_data.wait = 0;
  harness_forker_data.exit = 0;
  /* Initialize process table. */
  harness_forker_data.count = 0;
  harness_forker_data.procs = NULL;
  /* Initialize signal masks. */
  if ((-1 == sigemptyset(&sigmask                     ))||
      (-1 == sigemptyset(&harness_forker_data.sigmask))) {
    HARNESS_LOG((HARNESS_WARN(unable to empty signal mask)))
    return -1;
  }
  if ((-1 == sigaddset(&sigmask                     , SIGCHLD))||
      (-1 == sigaddset(&sigmask                     , SIGTERM))||
      (-1 == sigaddset(&sigmask                     , SIGPIPE))||
      (-1 == sigaddset(&harness_forker_data.sigmask, SIGPIPE))) {
    HARNESS_LOG((HARNESS_WARN(unable to add signal to signal mask)))
    return -1;
  }
  /* Block SIGCLD, SIGTERM and SIGPIPE signals. */
  if (-1 == pthread_sigmask(SIG_BLOCK, &sigmask, NULL)) {
    HARNESS_LOG((HARNESS_WARN(unable to set thread signal mask)))
    return -1;
  }
  /* Initialize wake-up pipe. */
  if (0 != pipe(harness_forker_data.wakeup)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize wake-up pipe)))
    return -1;
  }
  /* Set wake-up pipe non-blocking. */
  if (0 != fcntl(harness_forker_data.wakeup[0], F_SETFL, O_NONBLOCK)) {
    HARNESS_LOG((HARNESS_WARN(unable to set wake-up pipe to non-blocking)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_forker_data.wakeup[0]))||
        (0 != close(harness_forker_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Allocate buffer. */
  if (NULL == (harness_forker_data.buffer =
               malloc(harness_forker_data.size))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate buffer)))
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_forker_data.wakeup[0]))||
        (0 != close(harness_forker_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  /* Initialize the default envc. */
  for (harness_forker_data.envc = 0;
       NULL != environ[harness_forker_data.envc];
       harness_forker_data.envc++);
  /* Install signal handler for SIGCHLD and SIGTERM. */
  if ((SIG_ERR == signal(SIGTERM, harness_forker_handler))||
      (SIG_ERR == signal(SIGCHLD, harness_forker_handler))) {
    HARNESS_LOG((HARNESS_WARN(unable to install signal handler)))
    /* Deallocate buffer. */
    free(harness_forker_data.buffer);
    harness_forker_data.buffer = NULL;
    /* Finalize wake-up pipe. */
    if ((0 != close(harness_forker_data.wakeup[0]))||
        (0 != close(harness_forker_data.wakeup[1]))) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    }
    return -1;
  }
  return 0;
}

/*
 * Finalizes the forker process.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_fini"
int harness_forker_fini () {
  int result;
  result = 0;
  HARNESS_LOG((HARNESS_INFO(forker process shutdown)))
  /* Reset signal handler for SIGCHLD and SIGTERM. */
  if ((SIG_ERR == signal(SIGTERM, SIG_IGN))||
      (SIG_ERR == signal(SIGCHLD, SIG_IGN))) {
    HARNESS_LOG((HARNESS_WARN(unable to reset signal handler)))
    result = -1;
  }
  /* Finalize the default envc. */
  harness_forker_data.envc = 0;
  /* Deallocate buffer. */
  free(harness_forker_data.buffer);
  harness_forker_data.buffer = NULL;
  /* Finalize wake-up pipe. */
  if ((0 != close(harness_forker_data.wakeup[0]))||
      (0 != close(harness_forker_data.wakeup[1]))) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize wake-up pipe)))
    result = -1;
  }
  harness_forker_data.wakeup[0] = -1;
  harness_forker_data.wakeup[1] = -1;
  /* Finalize process table. */
  harness_forker_data.count = 0;
  if (NULL != harness_forker_data.procs) {
    free(harness_forker_data.procs);
    harness_forker_data.procs = NULL;
  }
  /* Reset wait and exit flags. */
  harness_forker_data.wait = 0;
  harness_forker_data.exit = 0;
  /* Reset I/O pipe ends. */
  harness_forker_data.pipes[0] = -1;
  harness_forker_data.pipes[1] = -1;
  return result;
}

/*
 * The signal handler of the forker process.
 *
 * signum = The number of the received signal.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_handler"
void harness_forker_handler (int signum) {
  /* Check the number of the received signal. */
  switch (signum) {
    /* Check for SIGCHLD signal. */
    case SIGCHLD: {
      /* Set wait flag. */
      harness_forker_data.wait = 1;
      /* Send byte to wake-up pipe. */
      if (0 != harness_forker_write(harness_forker_data.wakeup[1], 1, "")) {
        HARNESS_LOG((HARNESS_WARN(unable to send byte to wake-up pipe)))
        exit(EXIT_FAILURE);
      }
      /* Reinstall signal handler for SIGCHLD. */
      if (SIG_ERR == signal(SIGCHLD, harness_forker_handler)) {
        HARNESS_LOG((HARNESS_WARN(unable to reinstall signal handler)))
        exit(EXIT_FAILURE);
      }
      break;
    }
    /* Check for SIGTERM signal. */
    case SIGTERM: {
      HARNESS_LOG((
        HARNESS_INFO(forker process shutdown by signal %s), "SIGTERM"))
      /* Ckeck exit flag. */
      if (1 != harness_forker_data.exit) {
        /* Set exit flag. */
        harness_forker_data.exit = 1;
        /* Send byte to wake-up pipe. */
        if (0 != harness_forker_write(harness_forker_data.wakeup[1], 1, "")) {
          HARNESS_LOG((HARNESS_WARN(unable to send byte to wake-up pipe)))
          exit(EXIT_FAILURE);
        }
      }
      /* Reinstall signal handler for SIGTERM. */
      if (SIG_ERR == signal(SIGTERM, harness_forker_handler)) {
        HARNESS_LOG((HARNESS_WARN(unable to reinstall signal handler)))
        exit(EXIT_FAILURE);
      }
      break;
    }
    /* Check for any other signal. */
    default: {
      HARNESS_LOG((
        HARNESS_INFO(forker process shutdown by signal %d), signum))
      /* Set exit flag. */
      harness_forker_data.exit = 1;
      /* Send byte to wake-up pipe. */
      if (0 != harness_forker_write(harness_forker_data.wakeup[1], 1, "")) {
        HARNESS_LOG((HARNESS_WARN(unable to send byte to wake-up pipe)))
        exit(EXIT_FAILURE);
      }
      break;
    }
  }
}

/*
 * Runs the forker process tasks.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_run"
int harness_forker_run () {
  int          maxfd, fds;
  fd_set       fdset;
  unsigned int index;
  HARNESS_LOG((HARNESS_INFO(forker process running)))
  while (1) {
    /* Empty file descriptor set. */
    FD_ZERO(&fdset);
    /* Add incoming I/O pipe and wake-up pipe to file descriptor set. */
    FD_SET(harness_forker_data.wakeup[0], &fdset);
    FD_SET(harness_forker_data.pipes[0],  &fdset);
    /* Calculate the greatest file descriptor in the set. */
    maxfd = harness_forker_data.wakeup[0];
    if (maxfd < harness_forker_data.pipes[0]) {
      maxfd = harness_forker_data.pipes[0];
    }
    /* Add pipes from all child processes to file descriptor set. */
    for (index = 0; index < harness_forker_data.count; index++) {
      /* Ckeck if pipe from child process is open. */
      if (-1 != harness_forker_data.procs[index].pipes[0]) {
        /* Add pipe from child process to file descriptor set. */
        FD_SET(harness_forker_data.procs[index].pipes[0],  &fdset);
        /* Calculate the greatest file descriptor in the set. */
        if (maxfd < harness_forker_data.procs[index].pipes[0]) {
          maxfd = harness_forker_data.procs[index].pipes[0];
        }
      }
    }
    /* Wait on file descriptor set. */
    fds = pselect(maxfd + 1, &fdset, NULL, NULL, NULL,
                  &harness_forker_data.sigmask);
    /* Ckeck for error. */
    if ((-1 == fds)&&(EINTR != errno)) {
      HARNESS_LOG((HARNESS_WARN(unable to wait on file descriptor set)))
      return -1;
    }
    /* Ckeck exit flag. */
    if (1 == harness_forker_data.exit) {
      int result;
      /* Send SIGTERM to kernel process. */
      if (0 != kill(getppid(), SIGTERM)) {
        HARNESS_LOG((HARNESS_WARN(unable to send sigterm to kernel process)))
      }
      /* Empty incoming I/O pipe until it is closed. */
      while (0 <= (result =
                   harness_forker_read(harness_forker_data.pipes[0],
                                       harness_forker_data.size,
                                       harness_forker_data.buffer, 1)));
      if (-2 != result) {
        HARNESS_LOG((HARNESS_WARN(unable to empty incoming I/O pipe)))
      }
      return 0;
    }
    /* Ckeck ready wake-up pipe. */
    if ((0 < fds)&&(0 != FD_ISSET(harness_forker_data.wakeup[0], &fdset))) {
      /* Reduce the number of ready file descriptors. */
      fds--;
      /* Try to empty wake-up pipe. */
      if (0 > harness_forker_read(harness_forker_data.wakeup[0],
                                  harness_forker_data.size,
                                  harness_forker_data.buffer, 0)) {
        HARNESS_LOG((HARNESS_WARN(unable to empty wake-up pipe)))
        return -1;
      }
    }
    /* Ckeck for ready pipes from child processes. */
    for (index = 0; 0 < fds && index < harness_forker_data.count; index++) {
      /* Ckeck if pipe from child process is open and ready. */
      if ((-1 != harness_forker_data.procs[index].pipes[0]) &&
          (0  != FD_ISSET(harness_forker_data.procs[index].pipes[0],
                          &fdset))) {
        /* Reduce the number of ready file descriptors. */
        fds--;
        /* Forward child process output to kernel. */
        if (0 != harness_forker_forward(index)) {
          HARNESS_LOG((
            HARNESS_WARN(unable to forward child process output to kernel)))
          return -1;
        }
        /* Ckeck exit flag. */
        if (1 == harness_forker_data.exit) {
          return 0;
        }
      }
    }
    /* Ckeck wait flag. */
    if (1 == harness_forker_data.wait) {
      /* Reset wait flag. */
      harness_forker_data.wait = 0;
      /* Wait for exited child processes. */
      if (0 != harness_forker_wait()) {
        HARNESS_LOG((
          HARNESS_WARN(unable to wait for exited child processes)))
        return -1;
      }
      /* Ckeck exit flag. */
      if (1 == harness_forker_data.exit) {
        return 0;
      }
    }
    /* Ckeck for ready incoming I/O pipe. */
    if ((0 < fds)&&(0 != FD_ISSET(harness_forker_data.pipes[0], &fdset))) {
      /* Reduce the number of ready file descriptors. */
      fds--;
      /* Process kernel message. */
      if (0 != harness_forker_process()) {
        HARNESS_LOG((HARNESS_WARN(unable to process kernel message)))
        return -1;
      }
      /* Ckeck exit flag. */
      if (1 == harness_forker_data.exit) {
        return 0;
      }
    }
  }
  return 0;
}

/*
 * Processes kernel messages.
 *
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_process"
int harness_forker_process () {
  char command;
  /* Read command from incoming I/O pipe. */
  if (0 > harness_forker_read(harness_forker_data.pipes[0],
                               sizeof(char), &command, 1)) {
    HARNESS_LOG((HARNESS_INFO(forker process shutdown by closed pipe)))
    harness_forker_data.exit = 1;
    return 0;
  }
  /* Check command char. */
  switch (command) {
    /* Check for execution request. */
    case 'E': {
      char         **argv, **envv;
      unsigned int   handle, argc, envc;
      /* Read execution request. */
      if (0 != harness_forker_read_execute(&handle, &argc, &envc,
                                            &argv, &envv)) {
        HARNESS_LOG((HARNESS_WARN(unable to read execution request)))
        return -1;
      }
      /* Process execution request. */
      if (0 != harness_forker_process_execute(handle, argc, envc,
                                               argv, envv)) {
        HARNESS_LOG((HARNESS_WARN(unable to process execution request)))
        return -1;
      }
      return 0;
    }
    /* Check for input request. */
    case 'I': {
      unsigned int handle, length;
      /* Read input request. */
      if (0 != harness_forker_read_input(&handle, &length)) {
        HARNESS_LOG((HARNESS_WARN(unable to read input request)))
        return -1;
      }
      /* Process input request. */
      if (0 != harness_forker_process_input(handle, length)) {
        HARNESS_LOG((HARNESS_WARN(unable to process input request)))
        return -1;
      }
      return 0;
    }
    /* Check for unknown command */
    default: {
      HARNESS_LOG((HARNESS_WARN(unknown command from incoming I/O pipe)))
      return -1;
    }
  }
}

/*
 * Reads an execution request.
 *
 * handle = The process handle (return).
 * argc   = The argument vector count (return).
 * envc   = The environment vector count (return).
 * argv   = The argument vector (argument[0] is the program name) (return).
 * envv   = The environment vector (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_read_execute"
int harness_forker_read_execute (unsigned int   *handle,
                                 unsigned int   *argc,
                                 unsigned int   *envc,
                                 char         ***argv,
                                 char         ***envv) {
  unsigned int index, length;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check argc parameter. */
  if (NULL == argc) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argc parameter is null)))
    return -1;
  }
  /* Check envc parameter. */
  if (NULL == envc) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(envc parameter is null)))
    return -1;
  }
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check envc parameter. */
  if (NULL == envc) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(envc parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Read handle from incoming I/O pipe. */
  if (0 > harness_forker_read(harness_forker_data.pipes[0],
                              sizeof(unsigned int), handle, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read handle from incoming I/O pipe)))
    return -1;
  }
  /* Read argc from incoming I/O pipe. */
  if (0 > harness_forker_read(harness_forker_data.pipes[0],
                              sizeof(unsigned int), argc, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read argc from incoming I/O pipe)))
    return -1;
  }
  /* Read envc from incoming I/O pipe. */
  if (0 > harness_forker_read(harness_forker_data.pipes[0],
                              sizeof(unsigned int), envc, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read envc from incoming I/O pipe)))
    return -1;
  }
  *envc += harness_forker_data.envc;
  /* Allocate argv. */
  if (NULL == (*argv = malloc(((*argc) + 1) * sizeof(char*)))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate argv)))
    *argc = 0;
    return -1;
  }
  /* Allocate envv. */
  if (NULL == (*envv = malloc(((*envc) + 1) * sizeof(char*)))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate envv)))
    /* Deallocate argv. */
    free(*argv);
    return -1;
  }
  /* Read argv from incoming I/O pipe. */
  for (index = 0; index < (*argc); index++) {
    /* Read length from incoming I/O pipe. */
    if (0 > harness_forker_read(harness_forker_data.pipes[0],
                                sizeof(unsigned int), &length, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read length from incoming I/O pipe)))
      /* Deallocate envv. */
      free(*envv);
      /* Deallocate argv. */
      while (0 != index) {
        index--;
        free((*argv)[index]);
      }
      free(*argv);
      return -1;
    }
    /* Allocate argv string. */
    if (NULL == ((*argv)[index] = malloc(length))) {
      HARNESS_LOG((HARNESS_WARN(unable to allocate argv string)))
      /* Deallocate envv. */
      free(*envv);
      /* Deallocate argv. */
      while (0 != index) {
        index--;
        free((*argv)[index]);
      }
      free(*argv);
      return -1;
    }
    /* Read argv string from incoming I/O pipe. */
    if (0 > harness_forker_read(harness_forker_data.pipes[0], length,
                                (*argv)[index], 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read argv string from incoming I/O pipe)))
      /* Deallocate envv. */
      free(*envv);
      /* Deallocate argv. */
      free((*argv)[index]);
      while (0 != index) {
        index--;
        free((*argv)[index]);
      }
      free(*argv);
      return -1;
    }
  }
  (*argv)[index] = NULL;
  /* Initialize envv. */
  for (index = 0; index < harness_forker_data.envc; index++) {
    (*envv)[index] = environ[index];
  }
  /* Read envv from incoming I/O pipe. */
  for (; index < (*envc); index++) {
    /* Read length from incoming I/O pipe. */
    if (0 > harness_forker_read(harness_forker_data.pipes[0],
                                sizeof(unsigned int), &length, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read length from incoming I/O pipe)))
      /* Deallocate envv. */
      while (harness_forker_data.envc != index) {
        index--;
        free((*envv)[index]);
      }
      free(*envv);
      /* Deallocate argv. */
      for (index = 0; index < (*argc); index++) {
        free((*argv)[index]);
      }
      free(*argv);
      return -1;
    }
    /* Allocate envv string. */
    if (NULL == ((*envv)[index] = malloc(length))) {
      HARNESS_LOG((HARNESS_WARN(unable to allocate envv string)))
      /* Deallocate envv. */
      while (harness_forker_data.envc != index) {
        index--;
        free((*envv)[index]);
      }
      free(*envv);
      /* Deallocate argv. */
      for (index = 0; index < (*argc); index++) {
        free((*argv)[index]);
      }
      free(*argv);
      return -1;
    }
    /* Read envv string from incoming I/O pipe. */
    if (0 > harness_forker_read(harness_forker_data.pipes[0], length,
                                (*envv)[index], 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read envv string from incoming I/O pipe)))
      /* Deallocate envv. */
      free((*envv)[index]);
      while (harness_forker_data.envc != index) {
        index--;
        free((*envv)[index]);
      }
      free(*envv);
      /* Deallocate argv. */
      for (index = 0; index < (*argc); index++) {
        free((*argv)[index]);
      }
      free(*argv);
      return -1;
    }
  }
  (*envv)[index] = NULL;
  return 0;
}

/*
 * Processes an execution request.
 *
 * handle = The process handle.
 * argc   = The argument vector count.
 * envc   = The environment vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envv   = The environment vector.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_process_execute"
int harness_forker_process_execute (unsigned int   handle,
                                    unsigned int   argc,
                                    unsigned int   envc,
                                    char         **argv,
                                    char         **envv) {
  int           pipes[2][2];
  void         *procs;
  unsigned int  index;
  /* Create pipe one. */
  if (0 != pipe(pipes[0])) {
    HARNESS_LOG((HARNESS_WARN(unable to create pipe one)))
    return -1;
  }
  /* Create pipe two. */
  if (0 != pipe(pipes[1])) {
    HARNESS_LOG((HARNESS_WARN(unable to create pipe two)))
    /* Close pipe one. */
    close(pipes[0][0]);
    close(pipes[0][1]);
    return -1;
  }
  /* Set pipe one output to non-blocking. */
  if (0 != fcntl(pipes[0][0], F_SETFL, O_NONBLOCK)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to set pipe one output to non-blocking)))
    /* Close pipes one and two. */
    close(pipes[0][0]);
    close(pipes[0][1]);
    close(pipes[1][0]);
    close(pipes[1][1]);
    return -1;
  }
  /* Increase process table. */
  index = harness_forker_data.count;
  harness_forker_data.count++;
  /* Reallocate process table. */
  if (NULL == (procs = realloc(harness_forker_data.procs,
                               harness_forker_data.count *
                        sizeof(harness_forker_data.procs[0])))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate process table)))
    /* Close pipes one and two. */
    close(pipes[0][0]);
    close(pipes[0][1]);
    close(pipes[1][0]);
    close(pipes[1][1]);
    return -1;
  }
  harness_forker_data.procs = procs;
  /* Create process. */
  switch (harness_forker_data.procs[index].pid = fork()) {
    /* Check for error. */
    case -1: {
      HARNESS_LOG((HARNESS_WARN(unable to fork process)))
      close(pipes[0][0]);
      close(pipes[0][1]);
      close(pipes[1][0]);
      close(pipes[1][1]);
      return -1;
    }
    /* Check for child process. */
    case 0: {
      /* Close pipe one output and pipe two input. */
      close(pipes[0][0]);
      close(pipes[1][1]);
      /* Duplicate pipe one input to stdout. */
      if (-1 == dup2(pipes[0][1], 1)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to duplicate pipe one pipe end to stdio)))
        /* Close other pipe ends. */
        close(pipes[0][1]);
        close(pipes[1][0]);
        /* Exit with failure. */
        exit(EXIT_FAILURE);
      }
      /* Duplicate pipe two output to stdin. */
      if (-1 == dup2(pipes[1][0], 0)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to duplicate pipe two pipe end to stdio)))
        /* Close other pipe ends. */
        close(pipes[0][1]);
        close(pipes[1][0]);
        /* Exit with failure. */
        exit(EXIT_FAILURE);
      }
      /* Close pipe one input and pipe two output. */
      close(pipes[0][1]);
      close(pipes[1][0]);
      /* Set envv of new process. */
      environ = envv;
      /* Execute program. */
      execv(argv[0], argv);
      HARNESS_LOG((HARNESS_WARN(unable to execute program)))
      /* Exit with failure. */
      exit(EXIT_FAILURE);
    }
    /* Check for parent process. */
    default: {
      /* Close pipe one input and pipe two output. */
      close(pipes[0][1]);
      close(pipes[1][0]);
      /* Save process table entry. */
      harness_forker_data.procs[index].handle   = handle;
      harness_forker_data.procs[index].pipes[0] = pipes[0][0];
      harness_forker_data.procs[index].pipes[1] = pipes[1][1];
      /* Write command to outgoing I/O pipe. */
      if (0 != harness_forker_write(harness_forker_data.pipes[1], 1, "e")) {
        HARNESS_LOG((HARNESS_INFO(forker process shutdown by closed pipe)))
        harness_forker_data.exit = 1;
        return 0;
      }
      /* Write process handle to outgoing I/O pipe. */
      if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                    sizeof(unsigned int), &handle)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to write process handle to outgoing I/O pipe)))
        return -1;
      }
      /* Write pid to outgoing I/O pipe. */
      if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                    sizeof(pid_t), 
                                    &harness_forker_data.procs[index].pid)) {
        HARNESS_LOG((
          HARNESS_WARN(unable to write pid to outgoing I/O pipe)))
        return -1;
      }
      return 0;
    }
  }
}

/*
 * Reads an input request.
 *
 * handle = The process handle (return).
 * length = The process input length (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_read_input"
int harness_forker_read_input (unsigned int *handle,
                               unsigned int *length) {
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    return -1;
  }
  /* Check length parameter. */
  if (NULL == length) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(length parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Read handle from incoming I/O pipe. */
  if (0 > harness_forker_read(harness_forker_data.pipes[0],
                              sizeof(unsigned int), handle, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read handle from incoming I/O pipe)))
    return -1;
  }
  /* Read length from incoming I/O pipe. */
  if (0 > harness_forker_read(harness_forker_data.pipes[0],
                              sizeof(unsigned int), length, 1)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to read length from incoming I/O pipe)))
    return -1;
  }
  /* Check length. */
  if (*length > harness_forker_data.size) {
    HARNESS_LOG((HARNESS_WARN(unable to read more than buffer size)))
    return -1;
  }
  /* Read buffer from incoming I/O pipe. */
  if (0 > harness_forker_read(harness_forker_data.pipes[0],
                              *length, harness_forker_data.buffer, 1)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to read buffer from incoming I/O pipe)))
    return -1;
  }
  return 0;
}

/*
 * Processes process input from the kernel.
 *
 * handle = The process handle (return).
 * length = The process input length (return).
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_process_input"
int harness_forker_process_input (unsigned int handle,
                                  unsigned int length) {
  int          fd;
  unsigned int index;
  /* Search for process table entry. */
  for (index = 0; index < harness_forker_data.count; index++) {
    if (handle == harness_forker_data.procs[index].handle) {
      break;
    }
  }
  /* Check for process table entry. */
  if (index == harness_forker_data.count) {
    HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
    return 0;
  }
  /* Check for closed outgoing process pipe. */
  if (-1 == harness_forker_data.procs[index].pipes[1]) {
    HARNESS_LOG((HARNESS_INFO(unable to write to closed pipe)))
    return 0;
  }
  /* Write buffer to outgoing process pipe. */
  if (0 != harness_forker_write(harness_forker_data.procs[index].pipes[1],
                                length, harness_forker_data.buffer)) {
    HARNESS_LOG((HARNESS_INFO(unable to write to closed pipe)))
    /* Reset outgoing process pipe. */
    harness_forker_data.procs[index].pipes[1] = -1;
    /* Write command to outgoing I/O pipe. */
    if (0 != harness_forker_write(harness_forker_data.pipes[1], 1, "p")) {
      HARNESS_LOG((HARNESS_INFO(forker process shutdown by closed pipe)))
      harness_forker_data.exit = 1;
      return 0;
    }
    /* Write process handle to outgoing I/O pipe. */
    if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                  sizeof(unsigned int), 
                                 &harness_forker_data.procs[index].handle)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write process handle to outgoing I/O pipe)))
      return -1;
    }
    /* Write closed pipe to outgoing I/O pipe. */
    fd = 1;
    if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                  sizeof(int), &fd)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write closed pipe to outgoing I/O pipe)))
      return -1;
    }
  }
  return 0;
}

/*
 * Forwards process output to the kernel.
 *
 * length = The index in the forker process table.
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_forward"
int harness_forker_forward (unsigned int index) {
  int length, fd;
  /* Read some process output into buffer. */
  if (0 > (length = harness_forker_read(
                     harness_forker_data.procs[index].pipes[0],
                     harness_forker_data.size,
                     harness_forker_data.buffer, 0))) {
    /* Reset incoming process pipe. */
    harness_forker_data.procs[index].pipes[0] = -1;
    /* Write command to outgoing I/O pipe. */
    if (0 != harness_forker_write(harness_forker_data.pipes[1], 1, "p")) {
      HARNESS_LOG((HARNESS_INFO(forker process shutdown by closed pipe)))
      harness_forker_data.exit = 1;
      return 0;
    }
    /* Write process handle to outgoing I/O pipe. */
    if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                  sizeof(unsigned int), 
                                 &harness_forker_data.procs[index].handle)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write process handle to outgoing I/O pipe)))
      return -1;
    }
    /* Write closed pipe to outgoing I/O pipe. */
    fd = 0;
    if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                  sizeof(int), &fd)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write closed pipe to outgoing I/O pipe)))
      return -1;
    }
    return 0;
  }
  /* Write command to outgoing I/O pipe. */
  if (0 != harness_forker_write(harness_forker_data.pipes[1], 1, "o")) {
    HARNESS_LOG((HARNESS_INFO(forker process shutdown by closed pipe)))
    harness_forker_data.exit = 1;
    return 0;
  }
  /* Write process handle to outgoing I/O pipe. */
  if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                sizeof(unsigned int),
                                &harness_forker_data.procs[index].handle)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to write process handle to outgoing I/O pipe)))
    return -1;
  }
  /* Write length to outgoing I/O pipe. */
  if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                sizeof(unsigned int), &length)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to write length to outgoing I/O pipe)))
    return -1;
  }
  /* Write buffer to outgoing I/O pipe. */
  if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                length, harness_forker_data.buffer)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to write length to outgoing I/O pipe)))
    return -1;
  }
  return 0;
}

/*
 * Waits for exited child processes..
 *
 * return = 0 on success, -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_wait"
int harness_forker_wait () {
  int           status;
  pid_t         pid;
  void         *procs;
  unsigned int  index, handle;
  /* Wait (non-blocking) for exited processes while there are any. */
  while (1) {
    /* Wait (non-blocking) for any exited process. */
    if (0 >= (pid = waitpid(-1, &status, WNOHANG))) {
      return 0;
    }
    /* Search for forker process table entry. */
    for (index = 0; index < harness_forker_data.count; index++) {
      if (pid == harness_forker_data.procs[index].pid) {
        break;
      }
    }
    /* Check for forker process table entry. */
    if (index == harness_forker_data.count) {
      HARNESS_LOG((HARNESS_INFO(unable to find process table entry)))
      return -1;
    }
    /* Save process handle. */
    handle = harness_forker_data.procs[index].handle;
    /* Remove process table entry. */
    harness_forker_data.count--;
    memmove(harness_forker_data.procs + index,
            harness_forker_data.procs + index + 1,
           (harness_forker_data.count - index) *
     sizeof(harness_forker_data.procs[0]));
    /* Reallocate process table. */
    if (0 == harness_forker_data.count) {
      free(harness_forker_data.procs);
      harness_forker_data.procs = NULL;
    } else if (NULL == (procs = realloc(harness_forker_data.procs,
                                        harness_forker_data.count *
                                 sizeof(harness_forker_data.procs[0])))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate process table entry)))
    } else {
      harness_forker_data.procs = procs;
    }
    /* Write command to outgoing I/O pipe. */
    if (0 != harness_forker_write(harness_forker_data.pipes[1], 1, "q")) {
      HARNESS_LOG((HARNESS_INFO(forker process shutdown by closed pipe)))
      harness_forker_data.exit = 1;
      return 0;
    }
    /* Write process handle to outgoing I/O pipe. */
    if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                  sizeof(unsigned int), &handle)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write process handle to outgoing I/O pipe)))
      return -1;
    }
    /* Write exit status to outgoing I/O pipe. */
    if (0 != harness_forker_write(harness_forker_data.pipes[1],
                                  sizeof(int), &status)) {
      HARNESS_LOG((
        HARNESS_WARN(unable to write exit status to outgoing I/O pipe)))
      return -1;
    }
  }
}

/*
 * Reads a buffer from a file descriptor (non-/blocking).
 *
 * fd     = The file descriptor to read from.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed file descriptor or -1
 *          on any other error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_read"
int harness_forker_read (int           fd,
                         unsigned int  length,
                         void         *buffer,
                         unsigned int  block) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; index < length;) {
    /* Read some data. */
    switch (bytes = read(fd, buffer + index, length - index)) {
      case -1: {
        switch (errno) {
          case EINTR: {
            break;
          }
          case EAGAIN: {
            if (0 == block) {
              return index;
            }
            break;
          }
          default: {
            HARNESS_LOG((
              HARNESS_WARN(unable to read from file descriptor)))
            return -1;
          }
        }
        break;
      }
      case 0: {
        errno = EPIPE;
        if (0 != index) {
          HARNESS_LOG((
            HARNESS_WARN(unable to read from closed file descriptor)))
        }
        return -2;
      }
      default: {
        index += bytes;
        if (0 == block) {
          return index;
        }
      }
    }
  }
  return index;
}

/*
 * Writes a buffer into a file descriptor (blocking).
 *
 * fd     = The file descriptor to write to.
 * length = The buffer length to write.
 * buffer = The buffer to write from.
 * return = 0 on success, -2 on closed file descriptor or -1 on any other error
 *          with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_write"
int harness_forker_write (int           fd,
                          unsigned int  length,
                          const void   *buffer) {
  int          bytes;
  unsigned int index;
#ifdef DEBUG
  /* Check length and buffer parameters. */
  if ((0 != length)&&(NULL == buffer)) {
    errno = EINVAL;
    HARNESS_LOG((
      HARNESS_WARN(buffer parameter is NULL and length parameter is not 0)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; index < length;) {
    /* Write some data. */
    switch (bytes = write(fd, buffer + index, length - index)) {
      case -1: {
        switch (errno) {
          case EINTR:
          case EAGAIN: {
            break;
          }
          case EPIPE: {
            if (0 != index) {
              HARNESS_LOG((
                HARNESS_WARN(unable to write to closed file descriptor)))
            }
            return -2;
          }
          default: {
            HARNESS_LOG((
              HARNESS_WARN(unable to write to file descriptor)))
            return -1;
          }
        }
        break;
      }
      default: {
        index += bytes;
      }
    }
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
