/******************************************************************************
 *
 * Source file for harness-example (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness-example.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Harness-example plug-in module data type. */
typedef struct {
  pthread_mutex_t  mutex;     /* mutex              */
  struct {
    unsigned int   count;     /* handle array count */
    unsigned int  *array;     /* handle array       */
  }                instances; /* instances          */
} harness_example_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* 
 * Initializes the harness-example plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_INIT(harness_example_init);

/*
 * Finalizes the harness-example plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_FINI(harness_example_fini);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Harness-example plug-in module data. */
harness_example_data_t harness_example_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex              */
  {
    0,                       /* handle array count */
    NULL                     /* handle array       */
  }                          /* instances          */
};

/* Harness-example plug-in "object". */
harness_example_t harness_example = {
  {
    LIBHARNESS_EXAMPLE_VERSION_CURRENT,  /* current version         */
    LIBHARNESS_EXAMPLE_VERSION_REVISION, /* current revision        */
    LIBHARNESS_EXAMPLE_VERSION_AGE,      /* version age             */
    LIBHARNESS_EXAMPLE_VERSION_FIRST     /* first supported version */
  },                                     /* plug-in version         */
  /**********************************************/
  /* PUT YOUR PUBLIC PLUG-IN DATA POINTERS HERE */
  /**********************************************/
  /**************************************************/
  /* PUT YOUR PUBLIC PLUG-IN FUNCTION POINTERS HERE */
  /**************************************************/
  harness_example_instances
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the harness-example plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_example_init"
HARNESS_PLUGINS_INIT(harness_example_init) {
  int           error;
  void         *array;
  unsigned int  index;
  HARNESS_LOG((HARNESS_INFO(harness-example startup)))
  /* Lock harness-example plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_example_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock harness-example plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in handle array. */
  for (index = 0; index < harness_example_data.instances.count; index++) {
    if (harness_example_data.instances.array[index] == handle) {
      errno = EALREADY;
      HARNESS_LOG((HARNESS_WARN(handle is already in handle array)))
      errno = EALREADY;
      harness_syserr();
      /* Unlock harness-example plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_example_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-example plug-in mutex)))
        errno = error;
        harness_syserr();
      }
      errno = EALREADY;
      return -1;
    }
  }
  /* Increase handle array. */
  index = harness_example_data.instances.count;
  harness_example_data.instances.count++;
  /* Reallocate handle array. */
  if (NULL == (array = realloc(harness_example_data.instances.array,
                               harness_example_data.instances.count *
                        sizeof(harness_example_data.instances.array[0])))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to reallocate handle array)))
    /* Decrease handle array. */
    harness_example_data.instances.count--;
    /* Unlock harness-example plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_example_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-example plug-in mutex)))
      errno = error;
      harness_syserr();
    }
    errno2 = errno;
    return -1;
  }
  harness_example_data.instances.array = array;
  /* Save handle array entry. */
  harness_example_data.instances.array[index] = handle;
  /* Check for first initialization. */
  if (1 == harness_example_data.instances.count) {
    /***************************/
    /* PUT YOUR INIT CODE HERE */
    /***************************/
  }
  /* Unlock harness-example plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_example_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-example plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the harness-example plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_example_fini"
HARNESS_PLUGINS_FINI(harness_example_fini) {
  int           error, result, errno2;
  void         *array;
  unsigned int  index;
  result = 0;
  errno2 = 0;
  HARNESS_LOG((HARNESS_INFO(harness-example shutdown)))
  /* Lock harness-example plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_example_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock harness-example plug-in mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Search for handle in handle array. */
  for (index = 0; index < harness_example_data.instances.count; index++) {
    if (harness_example_data.instances.array[index] == handle) {
      break;
    }
  }
  /* Check if handle is not in handle array. */
  if (index == harness_example_data.instances.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(handle is not in handle array)))
    errno = ESRCH;
    harness_syserr();
    /* Unlock harness-example plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_example_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-example plug-in mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Remove handle from handle array. */
  harness_example_data.instances.count--;
  memmove(harness_example_data.instances.array + index,
          harness_example_data.instances.array + index + 1,
         (harness_example_data.instances.count   - index) *
   sizeof(harness_example_data.instances.array[0]));
  /* Reallocate handle array. */
  if (0 == harness_example_data.instances.count) {
    free(harness_example_data.instances.array);
    harness_example_data.instances.array = NULL;
  } else if (NULL ==
             (array = realloc(harness_example_data.instances.array,
                              harness_example_data.instances.count *
                       sizeof(harness_example_data.instances.array[0])))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate handle array)))
    errno = 0;
  } else {
    harness_example_data.instances.array = array;
  }
  /* Check for last finalization. */
  if (0 == harness_example_data.instances.count) {
    /***************************/
    /* PUT YOUR FINI CODE HERE */
    /***************************/
  }
  /* Unlock harness-example plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_example_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-example plug-in mutex)))
    errno = error;
    harness_syserr();
    if (0 != errno2) {
      errno = errno2;
    } else {
      errno = error;
    }
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Returns the number of loaded instances of this harness-example plug-in
 * without locking its plug-in mutex.
 *
 * return = The number of loaded instances of this harness-example plug-in.
 */
#undef  __FUNC__
#define __FUNC__ "harness_example_instances"
unsigned int harness_example_instances () {
  return harness_example_data.instances.count;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
