#! /bin/sh
################################################################################
#
# Package autogen.sh script.
# Copyright (c) 2002-2006 Oak Ridge National Laboratory.
#
# For more information see the following files in the source distribution top-
# level directory or package data directory (usually /usr/local/share/package):
#
# - README    for general package information.
# - INSTALL   for package install information.
# - COPYING   for package license information and copying conditions.
# - AUTHORS   for package authors information.
# - ChangeLog for package changes information.
#
# Run `autogen.sh` to automatically and recursively create and execute
# `configure` scripts and their outputs.
#
################################################################################

# Set defaults for command line options.
ARGS=

# Parse command line options.
until test -z "$1"; do
  if test "$1" = "--help" || test "$1" = "-h"; then
    echo "Usage:"
    echo " autogen.sh [options]"
    echo "Options:"
    echo " -h, --help         Print out command line help."
    echo "     --noconfigure  Skip the configure process."
    echo " -v, --version      Print out version information."
    echo " Any other options are passed to configure."
    exit 0
  elif test "$1" = "--noconfigure"; then
    NOCONFIGURE=1
  elif test "$1" = "--version" || test "$1" = "-v"; then
    echo "autogen.sh 1.0"
    echo "Copyright (c) 2002-2006 Oak Ridge National Laboratory."
    echo "See COPYING in the source distribution top-level directory"
    echo "or package data directory (usually /usr/local/share/package)"
    echo "for license information and copying conditions."
    echo "Report bugs to engelmannc@ornl.gov."
    exit 0
  else
    ARGS="$ARGS $1"
  fi
  shift
done

# Recursively search for configuration files.
CONFIGS=`find -name configure.ac`

# Check if any configuration files where found.
if test -z "$CONFIGS"; then
  echo "checking for configs... none"
  echo
  echo "You must have at least one configure.ac file to compile this package."
  echo "Your source tarball appears incomplete."
  echo
  exit 1
else
  echo "checking for configs... found"
fi

# Check for `autoconf`.
if autoconf --version >/dev/null 2>&1; then
  echo "checking for autoconf... yes"
else
  echo "checking for autoconf... no"
  echo
  echo "You must have autoconf installed to compile this package. Download"
  echo "the appropriate package for your distribution, or get the source"
  echo "tarball at ftp://ftp.gnu.org/pub/gnu/."
  echo
  exit 1
fi

# Check for `automake` if needed.
if grep "^AM_INIT_AUTOMAKE" $CONFIGS >/dev/null; then
  echo "checking for automake... needed"
  RUN_AUTOMAKE=yes
  if automake --version >/dev/null 2>&1; then
    echo "checking for automake... yes"
  else
    echo "checking for automake... no"
    echo
    echo "You must have automake installed to compile this package. Download"
    echo "the appropriate package for your distribution, or get the source"
    echo "tarball at ftp://ftp.gnu.org/pub/gnu/."
    echo
    exit 1
  fi
else
  echo "checking for automake... not needed"
  RUN_AUTOMAKE=no
fi

# Check for `aclocal` if needed.
if test $RUN_AUTOMAKE = yes; then
  echo "checking for aclocal... needed"
  if aclocal --version >/dev/null 2>&1; then
    echo "checking for aclocal... yes"
  else
    echo "checking for aclocal... no"
    echo
    echo "The version of automake installed doesn\'t appear recent enough. You"
    echo "must have an automake version with aclocal installed to compile this"
    echo "package. Download the appropriate package for your distribution, or"
    echo "get the source tarball at ftp://ftp.gnu.org/pub/gnu/."
    echo
    exit 1
  fi
else
  echo "checking for aclocal... not needed"
fi

# Check for `autoheader` if needed.
if grep "^AM_CONFIG_HEADER" $CONFIGS >/dev/null; then
  echo "checking for autoheader... needed"
  RUN_AUTOHEADER=yes
  if autoheader --version >/dev/null 2>&1; then
    echo "checking for autoheader... yes"
  else
    echo "checking for autoheader... no"
    echo
    echo "You must have autoheader installed to compile this package. Download"
    echo "the appropriate package for your distribution, or get the source"
    echo "tarball at ftp://ftp.gnu.org/pub/gnu/."
    echo
    exit 1
  fi
else
  echo "checking for autoheader... not needed"
  RUN_AUTOHEADER=no
fi

# Check for `libtool` if needed.
if grep "^AC_PROG_LIBTOOL" $CONFIGS >/dev/null; then
  echo "checking for libtool... needed"
  RUN_LIBTOOL=yes
  if libtool --version >/dev/null 2>&1; then
    echo "checking for libtool... yes"
  else
    echo "checking for libtool... no"
    echo
    echo "You must have libtool installed to compile this package. Download"
    echo "the appropriate package for your distribution, or get the source"
    echo "tarball at ftp://ftp.gnu.org/pub/gnu/."
    echo
    exit 1
  fi
else
  echo "checking for libtool... not needed"
  RUN_LIBTOOL=no
fi

# Check for `gettext` if needed.
if grep "^AM_GNU_GETTEXT" $CONFIGS >/dev/null; then
  echo "checking for gettext... needed"
  if gettext --version >/dev/null 2>&1; then
    echo "checking for gettext... yes"
  else
    echo "checking for gettext... no"
    echo
    echo "You must have gettext installed to compile this package. Download"
    echo "the appropriate package for your distribution, or get the source"
    echo "tarball at ftp://ftp.gnu.org/pub/gnu/."
    echo
    exit 1
  fi
else
  echo "checking for gettext... not needed"
fi

# For every found configuration file, execute `libtool`, `automake`, `aclocal`
# and `autoheader` if needed. Always execute `autoconf`. Do not execute
# `configure` yet, since the `configure` script itself works recursively.
for CONFIG in $CONFIGS; do
  NEWDIR=`dirname $CONFIG`
  CONFIG=`basename $CONFIG`
  if test -f $NEWDIR/NOAUTOGEN; then
    echo "skipping $NEWDIR"
  else
    OLDDIR=`pwd`
    echo "entering $NEWDIR"
    cd $NEWDIR
    if grep "^AC_PROG_LIBTOOL" $CONFIG >/dev/null; then
      if grep "^AC_LIBLTDL" $CONFIG >/dev/null; then
        echo "running libtoolize --force --copy --ltdl"
        libtoolize --force --copy --ltdl
      else
        echo "running libtoolize --force --copy"
        libtoolize --force --copy
      fi
    else
      echo "skipping libtoolize"
    fi
    if grep "^AM_INIT_AUTOMAKE" $CONFIG >/dev/null; then
      if test -d m4; then
        echo "running aclocal -I m4"
        aclocal -I m4
      else
        echo "running aclocal"
        aclocal
      fi
    else
      echo "skipping aclocal"
    fi
    if grep "^AM_CONFIG_HEADER" $CONFIG >/dev/null; then
      echo "running autoheader"
      autoheader
    else
      echo "skipping autoheader"
    fi
    if grep "^AM_INIT_AUTOMAKE" $CONFIG >/dev/null; then
      echo "running automake --foreign --add-missing --copy"
      automake --foreign --add-missing --copy
    else
      echo "skipping automake"
    fi
    echo "running autoconf"
    autoconf
    echo "leaving $NEWDIR"
    cd $OLDDIR
  fi
done

# Recursively execute `configure` if desired.
# Use parameters of `autogen.sh` if any.
if test -z "$NOCONFIGURE"; then
  if test -z "$ARGS"; then
    echo
    echo "I am going to run configure with no arguments. If you wish to pass"
    echo "any to it, please specify them on the autogen.sh command line."
    echo
  fi
  echo "running configure $ARGS"
  ./configure $ARGS
else
  echo "skipping configure"
fi

################################################################################
#
# End of file.
#
################################################################################
