/******************************************************************************
 *
 * Source file for rmix localref (local object reference) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Creates a local object reference.
 *
 * localref  = The local object reference to create (return).
 * remoteref = The remote object reference.
 * provider  = The provider plug-in handle.
 * objectid  = The object id.
 * object    = The object (may be NULL).
 * interface = The object interface.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_localref_create"
int rmix_localref_create (rmix_localref_t        **localref,
                          const rmix_remoteref_t  *remoteref,
                          unsigned int             provider,
                          unsigned int             objectid,
                          void                    *object,
                          const rmix_interface_t  *interface) {
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate local object reference. */
  if (NULL == (*localref = malloc(sizeof(rmix_localref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate local object reference)))
    errno = errno2;
    return -1;
  }
  /* Duplicate remote object reference. */
  if (0 != rmix_remoteref_duplicate(&(*localref)->remoteref, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference)))
    /* Deallocate local object reference. */
    free(*localref);
    errno = errno2;
    return -1;
  }
  /* Initialize provider plug-in handle. */
  (*localref)->provider = provider;
  /* Initialize object id. */
  (*localref)->objectid = objectid;
  /* Initialize object. */
  (*localref)->object = object;
  /* Duplicate object interface. */
  if (0 != rmix_interface_duplicate(&(*localref)->interface, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object interface)))
    /* Destroy local object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&(*localref)->remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference duplicate)))
    }
    /* Deallocate local object reference. */
    free(*localref);
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a local object reference.
 *
 * localref   = The local object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding  (protocol specific) (may be NULL, "").
 * interfaces = The remote object interfaces list (NULL terminated or NULL).
 * provider   = The provider plug-in handle.
 * objectid   = The object id.
 * object     = The object (may be NULL).
 * interface  = The object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_localref_create2"
int rmix_localref_create2 (rmix_localref_t        **localref,
                           const char              *protocol,
                           const char              *binding,
                           const char             **interfaces,
                           const char              *location,
                           unsigned int             provider,
                           unsigned int             objectid,
                           void                    *object,
                           const rmix_interface_t  *interface) {
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate local object reference. */
  if (NULL == (*localref = malloc(sizeof(rmix_localref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate local object reference)))
    errno = errno2;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create2(&(*localref)->remoteref, protocol, binding,
                                  interfaces, location, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference)))
    /* Deallocate local object reference. */
    free(*localref);
    errno = errno2;
    return -1;
  }
  /* Initialize provider plug-in handle. */
  (*localref)->provider = provider;
  /* Initialize object id. */
  (*localref)->objectid = objectid;
  /* Initialize object. */
  (*localref)->object = object;
  /* Duplicate object interface. */
  if (0 != rmix_interface_duplicate(&(*localref)->interface, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object interface)))
    /* Destroy local object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&(*localref)->remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference duplicate)))
    }
    /* Deallocate local object reference. */
    free(*localref);
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a local object reference.
 *
 * localref   = The local object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding  (protocol specific) (may be NULL, "").
 * interfaces = The remote object interfaces list string (may be NULL).
 *              Individual interface names are delimited by ';'.
 * location   = The remote object location (protocol specific).
 * provider   = The provider plug-in handle.
 * objectid   = The object id.
 * object     = The object (may be NULL).
 * interface  = The object interface.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_localref_create3"
int rmix_localref_create3 (rmix_localref_t        **localref,
                           const char              *protocol,
                           const char              *binding,
                           const char              *interfaces,
                           const char              *location,
                           unsigned int             provider,
                           unsigned int             objectid,
                           void                    *object,
                           const rmix_interface_t  *interface) {
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate local object reference. */
  if (NULL == (*localref = malloc(sizeof(rmix_localref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate local object reference)))
    errno = errno2;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create3(&(*localref)->remoteref, protocol, binding,
                                  interfaces, location, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference)))
    /* Deallocate local object reference. */
    free(*localref);
    errno = errno2;
    return -1;
  }
  /* Initialize provider plug-in handle. */
  (*localref)->provider = provider;
  /* Initialize object id. */
  (*localref)->objectid = objectid;
  /* Initialize object. */
  (*localref)->object = object;
  /* Duplicate object interface. */
  if (0 != rmix_interface_duplicate(&(*localref)->interface, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object interface)))
    /* Destroy local object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&(*localref)->remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy local object reference duplicate)))
    }
    /* Deallocate local object reference. */
    free(*localref);
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Duplicates a local object reference.
 *
 * localref1 = The local object reference to create (return).
 * localref2 = The local object reference to duplicate.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_localref_duplicate"
int rmix_localref_duplicate (rmix_localref_t       **localref1,
                             const rmix_localref_t  *localref2) {
#ifdef DEBUG
  /* Check localref1 parameter. */
  if (NULL == localref1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localref2 parameter. */
  if (NULL == localref2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create local object reference duplicate. */
  if (0 != rmix_localref_create(localref1, localref2->remoteref,
                                localref2->provider, localref2->objectid,
                                localref2->object, localref2->interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create local object reference duplicate)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Destroys a local object reference.
 *
 * localref = The local object reference to destroy (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_localref_destroy"
int rmix_localref_destroy (rmix_localref_t **localref) {
  int result, errno2;
#ifdef DEBUG
  /* Check localref parameter. */
  if (NULL == localref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check *localref parameter. */
  if (NULL == *localref) {
    return 0;
  }
  result = 0;
  errno2 = 0;
  /* Destroy remote object reference duplicate. */
  if (0 != rmix_remoteref_destroy(&(*localref)->remoteref)) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    result = -1;
  }
  /* Destroy object interface duplicate. */
  if (0 != rmix_interface_destroy(&(*localref)->interface)) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy object interface duplicate)))
    result = -1;
  }
  /* Deallocate local object reference. */
  free(*localref);
  *localref = NULL;
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Compares two local object references.
 *
 * diff      = The difference between both local object references (return).
 *             diff > 0 if localref1 < localref2,
 *             diff < 0 if localref1 > localref2, and
 *             diff = 0 if localref1 = localref2
 * localref1 = The local object reference to compare against.
 * localref2 = The local object reference to compare with.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_localref_compare"
int rmix_localref_compare (int                   *diff,
                           const rmix_localref_t *localref1,
                           const rmix_localref_t *localref2) {
#ifdef DEBUG
  /* Check diff parameter. */
  if (NULL == diff) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(diff parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localref1 parameter. */
  if (NULL == localref1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check localref2 parameter. */
  if (NULL == localref2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(localref2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Compare and check for equal provider
     plug-in handles, object ids and objects. */
  if ((0 == (*diff = localref2->provider - localref1->provider))&&
      (0 == (*diff = localref2->objectid - localref1->objectid))&&
      (0 == (*diff = localref2->object   - localref1->object  ))) {
    /* Compare remote object references. */
    if (0 != rmix_remoteref_compare(diff, localref1->remoteref,
                                          localref2->remoteref)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to compare remote object references)))
      errno = errno2;
      return -1;
    }
    /* Check for equal remote object references. */
    if (0 == *diff) {
      /* Compare object interfaces. */
      if (0 != rmix_interface_compare(diff, localref1->interface,
                                            localref2->interface)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to compare object interfaces)))
        errno = errno2;
        return -1;
      }
    }
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
