/******************************************************************************
 *
 * Source file for the RMIX parameter (export parameter) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Creates an export parameter.
 *
 * parameter = The export parameter to create (return).
 * name      = The export parameter name (cannot be NULL).
 * value     = The export parameter value (cannot be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_parameter_create"
int rmix_parameter_create (rmix_parameter_t **parameter,
                           const char        *name,
                           const char        *value) {
#ifdef DEBUG
  /* Check parameter parameter. */
  if (NULL == parameter) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check value parameter. */
  if (NULL == value) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(value parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate export parameter. */
  if (NULL == (*parameter = malloc(sizeof(rmix_parameter_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate export parameter)))
    errno = errno2;
    return -1;
  }
  /* Duplicate export parameter name. */
  if (NULL == ((*parameter)->name = strdup(name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate export parameter name)))
    /* Deallocate export parameter. */
    free(*parameter);
    errno = errno2;
    return -1;
  }
  /* Duplicate export parameter value. */
  if (NULL == ((*parameter)->value = strdup(value))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate export parameter value)))
    /* Deallocate export parameter name. */
    free((*parameter)->name);
    /* Deallocate export parameter. */
    free(*parameter);
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Destroys an export parameter.
 *
 * parameter = The export parameter to destroy (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_parameter_destroy"
int rmix_parameter_destroy (rmix_parameter_t **parameter) {
#ifdef DEBUG
  /* Check parameter parameter. */
  if (NULL == parameter) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check parameter parameter. */
  if (NULL == *parameter) {
    return 0;
  }
  /* Deallocate export parameter value. */
  free((*parameter)->value);
  /* Deallocate export parameter name. */
  free((*parameter)->name);
  /* Deallocate export parameter. */
  free(*parameter);
  return 0;
}

/*
 * Duplicates an export parameter.
 *
 * parameter1 = The export parameter to create (return).
 * parameter2 = The export parameter to duplicate.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_parameter_duplicate"
int rmix_parameter_duplicate (rmix_parameter_t       **parameter1,
                              const rmix_parameter_t  *parameter2) {
#ifdef DEBUG
  /* Check parameter1 parameter. */
  if (NULL == parameter1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameter2 parameter. */
  if (NULL == parameter2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create export parameter. */
  if (0 != rmix_parameter_create(parameter1, parameter2->name,
                                 parameter2->value)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create export parameter)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Compares two export parameters.
 *
 * diff       = The difference between both export parameters (return).
 *              diff > 0 if parameter1 < parameter2,
 *              diff < 0 if parameter1 > parameter2, and
 *              diff = 0 if parameter1 = parameter2
 * parameter1 = The export parameter to compare against.
 * parameter2 = The export parameter to compare with.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_parameter_compare"
int rmix_parameter_compare (int                    *diff,
                            const rmix_parameter_t *parameter1,
                            const rmix_parameter_t *parameter2) {
#ifdef DEBUG
  /* Check diff parameter. */
  if (NULL == diff) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(diff parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameter1 parameter. */
  if (NULL == parameter1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameter1->name parameter. */
  if (NULL == parameter1->name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter1->name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameter1->value parameter. */
  if (NULL == parameter1->value) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter1->value parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameter2 parameter. */
  if (NULL == parameter2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameter2->name parameter. */
  if (NULL == parameter2->name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter2->name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameter2->value parameter. */
  if (NULL == parameter2->value) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameter2->value parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Compare export parameter names and check for equality. */
  if (0 != (*diff = strcmp(parameter1->name, parameter2->name))) {
    /* Compare export parameter values. */
    *diff = strcmp(parameter1->value, parameter2->value);
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
