/******************************************************************************
 *
 * Header file for rmix plugins (plug-in loader) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_PLUGINS_H
#define RMIX_PLUGINS_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Plug-in init and fini function signatures. */
#define RMIX_PLUGINS_INIT(name) int name (unsigned int handle)
#define RMIX_PLUGINS_FINI(name) int name (unsigned int handle)

/* Plug-in loading flags. */
#define RMIX_PLUGINS_UNIQUE 1 /* Load plug-in only once.          */
#define RMIX_PLUGINS_EXPORT 3 /* Export plug-in symbols globally. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Plug-in init and fini function data types. */
typedef RMIX_PLUGINS_INIT((*rmix_plugins_init_t));
typedef RMIX_PLUGINS_FINI((*rmix_plugins_fini_t));

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Loads a plug-in.
 *
 * handle = The plug-in handle (return). Plug-in handles are unique.
 * name   = The plug-in name. Plug-in names are unique and do
 *          not contain '/..' and do not start with '/'.
 * flags  = The plug-in loading flags, e.g. RMIX_PLUGINS_UNIQUE.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_load (unsigned int *handle,
                       const char   *name,
                       unsigned int  flags);

/*
 * Loads a plug-in.
 *
 * handle  = The plug-in handle (return). Plug-in handles are unique.
 * name    = The plug-in name. Plug-in names are unique and do
 *           not contain '/..' and do not start with '/'.
 * version = The version of the plug-in to load.
 * flags   = The plug-in loading flags, e.g. RMIX_PLUGINS_UNIQUE.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_load2 (unsigned int *handle,
                        const char   *name,
                        unsigned int  version,
                        unsigned int  flags);

/*
 * Loads a plug-in.
 *
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * first    = The first version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. RMIX_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_load3 (unsigned int *handle,
                        const char   *name,
                        unsigned int  first,
                        unsigned int  age,
                        unsigned int  revision,
                        unsigned int  flags);

/*
 * Loads a plug-in.
 *
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * current  = The current version supported of the plug-in to load.
 * age      = The number of previous versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. RMIX_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_load4 (unsigned int *handle,
                        const char   *name,
                        unsigned int  current,
                        unsigned int  age,
                        unsigned int  revision,
                        unsigned int  flags);

/*
 * Gets the module of a plug-in (for dlsym).
 *
 * module = The plug-in module (return).
 * handle = The plug-in handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_module (void         **module,
                         unsigned int   handle);

/*
 * Gets the symbol of a plug-in corresponding to a name (using dlsym).
 *
 * symbol = The plug-in symbol that corresponds to the name (return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_symbol (void         **symbol,
                         unsigned int   handle,
                         const char    *name);

/*
 * Gets the symbol of a plug-in corresponding to a name (using dlsym).
 * Prepends <plug-in name>_ (without any lib prefix) to the symbol name.
 *
 * symbol = The plug-in symbol that corresponds to the name (return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_symbol2 (void         **symbol,
                          unsigned int   handle,
                          const char    *name);

/*
 * Gets the name of a plug-in.
 *
 * name   = The plug-in name (return).
 * handle = The plug-in handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_name (const char   **name,
                       unsigned int   handle);

/*
 * Unloads a plug-in.
 *
 * handle = The plug-in handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_plugins_unload (unsigned int handle);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Plug-in loader directory. */
extern const char *rmix_plugins_dir;

#endif /* RMIX_PLUGINS_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
