/******************************************************************************
 *
 * Source file for rmix remoteref (remote object reference) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Remote object reference module data type. */
typedef struct {
  unsigned int host; /* host    identifier */
  unsigned int proc; /* process identifier */
  unsigned int rand; /* random  identifier */
} rmix_remoteref_data_t;

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Remote object reference module data. */
rmix_remoteref_data_t rmix_remoteref_data = {
  0, /* host    identifier */
  0, /* process identifier */
  0  /* random  identifier */
};

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the remote object reference module.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_init();

/*
 * Finalizes the remote object reference module.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_fini();

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the remote object reference module.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_init"
int rmix_remoteref_init() {
  struct timeval time;
  /* Initialize host identifier. */
  rmix_remoteref_data.host = 0;
  /* Initialize process identifier. */
  rmix_remoteref_data.proc = getpid();
  /* Get time of day for random seed. */
  if (0 != gettimeofday(&time, NULL)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get time of day)))
    errno = errno2;
    rmix_syserr();
    errno = errno2;
    return -1;
  }
  /* Seed random number generator. */
  srandom(time.tv_sec);
  /* Initialize random identifier. */
  rmix_remoteref_data.rand = random();
  return 0;
}

/*
 * Finalizes the remote object reference module.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_fini"
int rmix_remoteref_fini() {
  /* Finalize host, process and random identifier. */
  rmix_remoteref_data.host = 0;
  rmix_remoteref_data.proc = 0;
  rmix_remoteref_data.rand = 0;
  return 0;
}

/*
 * Creates a remote object reference.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding (protocol specific, may be NULL, "").
 * interfaces = The interfaces list (NULL terminated or NULL).
 * location   = The location (protocol specific).
 * object     = The object identifier.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create"
int rmix_remoteref_create (rmix_remoteref_t **remoteref,
                           const char        *protocol,
                           const char        *binding,
                           const char       **interfaces,
                           const char        *location,
                           const char        *object) {
  unsigned int index;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check location parameter. */
  if (NULL == location) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(location parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate remote object reference. */
  if (NULL == (*remoteref = malloc(sizeof(rmix_remoteref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Duplicate protocol. */
  if (NULL == ((*remoteref)->protocol = strdup(protocol))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate protocol)))
    /* Deallocate remote object reference. */
    free(*remoteref);
    errno = errno2;
    return -1;
  }
  /* Check binding parameter. */
  if (NULL != binding) {
    /* Duplicate binding. */
    if (NULL == ((*remoteref)->binding = strdup(binding))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate binding)))
      /* Deallocate protocol duplicate. */
      free((*remoteref)->protocol);
      /* Deallocate remote object reference. */
      free(*remoteref);
      errno = errno2;
      return -1;
    }
  } else {
    (*remoteref)->binding = NULL;
  }
  /* Check for empty interfaces list. */
  if ((NULL == interfaces)||(NULL == interfaces[0])) {
    /* Initialize interfaces list. */
    (*remoteref)->interfaces.count = 0;
    (*remoteref)->interfaces.array = NULL;
  } else {
    for (index = 0; NULL != interfaces[index]; index++);
    /* Allocate interfaces list array. */
    if (NULL == ((*remoteref)->interfaces.array =
                  malloc((index + 1) * sizeof(char*)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to allocate interfaces list array)))
      /* Deallocate binding duplicate. */
      free((*remoteref)->binding);
      /* Deallocate protocol duplicate. */
      free((*remoteref)->protocol);
      /* Deallocate remote object reference. */
      free(*remoteref);
      errno = errno2;
      return -1;
    }
    /* Duplicate interfaces list. */
    for (index = 0, (*remoteref)->interfaces.count = 0;
         NULL != interfaces[index];
         index++, (*remoteref)->interfaces.count++) {
      /* Duplicate interfaces list entry. */
      if (NULL == ((*remoteref)->interfaces.array[index] =
                    strdup(interfaces[index]))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to duplicate interfaces list entry)))
        /* Deallocate interfaces list entries. */
        for (index = 0; index < (*remoteref)->interfaces.count; index++) {
          /* Deallocate interfaces list entry. */
          free((*remoteref)->interfaces.array[index]);
        }
        /* Deallocate interfaces list array. */
        free((*remoteref)->interfaces.array);
        /* Deallocate binding duplicate. */
        free((*remoteref)->binding);
        /* Deallocate protocol duplicate. */
        free((*remoteref)->protocol);
        /* Deallocate remote object reference. */
        free(*remoteref);
        errno = errno2;
        return -1;
      }
    }
    /* Terminate interfaces list. */
    (*remoteref)->interfaces.array[index] = NULL;
  }
  /* Duplicate remote object reference location. */
  if (NULL == ((*remoteref)->location = strdup(location))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference location)))
    /* Deallocate interfaces list entries. */
    for (index = 0; index < (*remoteref)->interfaces.count; index++) {
      /* Deallocate interfaces list entry. */
      free((*remoteref)->interfaces.array[index]);
    }
    /* Deallocate interfaces list array. */
    if (NULL != (*remoteref)->interfaces.array) {
      free((*remoteref)->interfaces.array);
    }
    /* Deallocate binding duplicate. */
    free((*remoteref)->binding);
    /* Deallocate protocol duplicate. */
    free((*remoteref)->protocol);
    /* Deallocate remote object reference. */
    free(*remoteref);
    errno = errno2;
    return -1;
  }
  /* Duplicate object identifier. */
  if (NULL == ((*remoteref)->object = strdup(object))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object identifier)))
    /* Deallocate remote object reference location duplicate. */
    free((*remoteref)->location);
    /* Deallocate interfaces list entries. */
    for (index = 0; index < (*remoteref)->interfaces.count; index++) {
      /* Deallocate interfaces list entry. */
      free((*remoteref)->interfaces.array[index]);
    }
    /* Deallocate interfaces list array. */
    if (NULL != (*remoteref)->interfaces.array) {
      free((*remoteref)->interfaces.array);
    }
    /* Deallocate binding duplicate. */
    free((*remoteref)->binding);
    /* Deallocate protocol duplicate. */
    free((*remoteref)->protocol);
    /* Deallocate remote object reference. */
    free(*remoteref);
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding (protocol specific, may be NULL, "").
 * interfaces = The interfaces list (NULL terminated or NULL).
 * location   = The location (protocol specific).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create2"
int rmix_remoteref_create2 (rmix_remoteref_t **remoteref,
                            const char        *protocol,
                            const char        *binding,
                            const char       **interfaces,
                            const char        *location,
                            unsigned int       objectid) {
  char *object;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check location parameter. */
  if (NULL == location) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(location parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Construct object identifier. */
  if (-1 == asprintf(&object, "%u:%u:%u:%u", rmix_remoteref_data.host,
                     rmix_remoteref_data.proc, objectid,
                     rmix_remoteref_data.rand)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to construct object identifier)))
    errno2 = errno;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create(remoteref, protocol, binding, interfaces,
                                 location, object)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Deallocate object identifier. */
    free(object);
    errno2 = errno;
    return -1;
  }
  /* Deallocate object identifier. */
  free(object);
  return 0;
}

/*
 * Creates a remote object reference.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol.
 * binding    = The binding (protocol specific, may be NULL, "").
 * interfaces = The interfaces list string (may be NULL).
 *              Individual interface names are delimited by ';'.
 * interfaces = The interfaces list (NULL terminated or NULL).
 * location   = The location (protocol specific).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_remoteref_create3 (rmix_remoteref_t **remoteref,
                            const char        *protocol,
                            const char        *binding,
                            const char        *interfaces,
                            const char        *location,
                            unsigned int       objectid) {
  char         *object, **interfaces2;
  unsigned int  index1, index2, count;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check protocol parameter. */
  if (NULL == protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check location parameter. */
  if (NULL == location) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(location parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check for any interfaces. */
  if (NULL != interfaces) {
    /* Count the number of interfaces. */
    for (index1 = 0, count = 1; '\0' != interfaces[index1]; index1++) {
      if (';' == interfaces[index1]) {
        count++;
      }
    }
    /* Allocate interfaces list array. */
    if (NULL == (interfaces2 = malloc((count + 1) * sizeof(char*)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to allocate interfaces list array)))
      errno = errno2;
      return -1;
    }
    /* Terminate interfaces list array. */
    interfaces2[count] = NULL;
    /* Copy the interface names into the interfaces list array. */
    for (index1 = 0, index2 = 0, count = 0; 1; index2++) {
      if (('\0' == interfaces[index2])||
          (';'  == interfaces[index2])) {
        /* Allocate interface name. */
        if (NULL == (interfaces2[count] = malloc((index2 - index1 + 1) *
                                                 sizeof(char)))) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to allocate interface name)))
          /* Deallocate interface names. */
          while (0 != count--) {
            /* Deallocate interface name. */
            free(interfaces2[count]);
          }
          /* Deallocate interfaces list array. */
          free(interfaces2);
          errno = errno2;
          return -1;
        }
        /* Copy interface name. */
        memcpy(interfaces2[count], interfaces + index1, index2 - index1);
        interfaces2[count][index2 - index1] = '\0';
        /* Increase interface name count. */
        count++;
        /* Check for last interface name. */
        if ('\0' == interfaces[index2]) {
          break;
        }
        /* Reset next interface name start. */
        index1 = index2 + 1;
      }
    }
  } else {
    /* Set the number of interfaces to zero. */
    count       = 0;
    interfaces2 = NULL;
  }
  /* Construct object identifier. */
  if (-1 == asprintf(&object, "%u:%u:%u:%u", rmix_remoteref_data.host,
                     rmix_remoteref_data.proc, objectid,
                     rmix_remoteref_data.rand)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to construct object identifier)))
    /* Deallocate interface names. */
    while (0 != count--) {
      /* Deallocate interface name. */
      free(interfaces2[count]);
    }
    /* Deallocate interfaces list array. */
    if (NULL != interfaces2) {
      free(interfaces2);
    }
    errno = errno2;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create(remoteref, protocol, binding,
                                 (const char**)interfaces2,
                                 location, object)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Deallocate object identifier. */
    free(object);
    /* Deallocate interface names. */
    while (0 != count--) {
      /* Deallocate interface name. */
      free(interfaces2[count]);
    }
    /* Deallocate interfaces list array. */
    if (NULL != interfaces2) {
      free(interfaces2);
    }
    errno = errno2;
    return -1;
  }
  /* Deallocate object identifier. */
  free(object);
  /* Deallocate interface names. */
  while (0 != count--) {
    /* Deallocate interface name. */
    free(interfaces2[count]);
  }
  /* Deallocate interfaces list array. */
  if (NULL != interfaces2) {
    free(interfaces2);
  }
  return 0;
}

/*
 * Creates a remote object reference using the appropiate provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create4"
int rmix_remoteref_create4 (rmix_remoteref_t        **remoteref,
                            const char               *protocol,
                            const char               *binding,
                            const rmix_parameters_t  *parameters,
                            unsigned int              objectid) {
  void         *remoteref_create;
  unsigned int  provider;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ((NULL == protocol)||('\0' == protocol[0])) {
    /* Check parameters parameter. */
    if (NULL == parameters) {
      errno = EINVAL;
      RMIX_LOG((
        RMIX_WARN(parameters parameter is null and no protocol specified)))
      errno = EINVAL;
      return -1;
    }
    /* Search for additional PROTOCOL parameter. */
    if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
      if (ESRCH == errno) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(no protocol specified)))
        errno = EINVAL;
        return -1;
      } else {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
        errno = errno2;
        return -1;
      }
    }
  }
  /* Check binding parameter. */
  if (((NULL == binding)||('\0' == binding[0]))&&(NULL != parameters)) {
    /* Search for additional BINDING parameter. */
    if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
        (ESRCH != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in remoteref_create function. */
  if (0 != rmix_providers_symbol2(&remoteref_create, provider,
                                  "remoteref_create")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_create function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_create function. */
  if (NULL == remoteref_create) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_create function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != ((rmix_providers_remoteref_create_t)remoteref_create)
            (remoteref, protocol, binding, parameters, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Unload protocol provider plug-in. */
  if (0 != rmix_providers_unload(provider)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unload protocol provider plug-in)))
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference without loading the provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The preloaded provider plug-in handle.
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create4_p"
int rmix_remoteref_create4_p (rmix_remoteref_t        **remoteref,
                              unsigned int              provider,
                              const char               *protocol,
                              const char               *binding,
                              const rmix_parameters_t  *parameters,
                              unsigned int              objectid) {
  void *remoteref_create;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ((NULL == protocol)||('\0' == protocol[0])) {
    /* Check parameters parameter. */
    if (NULL == parameters) {
      errno = EINVAL;
      RMIX_LOG((
        RMIX_WARN(parameters parameter is null and no protocol specified)))
      errno = EINVAL;
      return -1;
    }
    /* Search for additional PROTOCOL parameter. */
    if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
      if (ESRCH == errno) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(no protocol specified)))
        errno = EINVAL;
        return -1;
      } else {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
        errno = errno2;
        return -1;
      }
    }
  }
  /* Check binding parameter. */
  if (((NULL == binding)||('\0' == binding[0]))&&(NULL != parameters)) {
    /* Search for additional BINDING parameter. */
    if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
        (ESRCH != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Resolve provider plug-in remoteref_create function. */
  if (0 != rmix_providers_symbol2(&remoteref_create, provider,
                                  "remoteref_create")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_create function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_create function. */
  if (NULL == remoteref_create) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_create function does not exist)))
    errno = ENOTSUP;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != ((rmix_providers_remoteref_create_t)remoteref_create)
            (remoteref, protocol, binding, parameters, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference using the appropiate provider plug-in and
 * without un loading the provider plug-in, while forcing a specific object
 * id. This is very useful for accessing standard object exports at the
 * client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The provider plug-in handle (return).
 * protocol   = The communication protocol. Overrides any additional PROTOCOL
 *              parameter. May only be NULL or "" if additional parameters
 *              contains PROTOCOL.
 * binding    = The binding  (protocol specific) (may be NULL, ""). Overrides
 *              any additional BINDING parameter.
 * parameters = Additional parameters (protocol specific) (may be NULL).
 * objectid   = The object id.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create4_p2"
int rmix_remoteref_create4_p2 (rmix_remoteref_t        **remoteref,
                               unsigned int             *provider,
                               const char               *protocol,
                               const char               *binding,
                               const rmix_parameters_t  *parameters,
                               unsigned int              objectid) {
  void         *remoteref_create;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check protocol parameter. */
  if ((NULL == protocol)||('\0' == protocol[0])) {
    /* Check parameters parameter. */
    if (NULL == parameters) {
      errno = EINVAL;
      RMIX_LOG((
        RMIX_WARN(parameters parameter is null and no protocol specified)))
      errno = EINVAL;
      return -1;
    }
    /* Search for additional PROTOCOL parameter. */
    if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
      if (ESRCH == errno) {
        errno = EINVAL;
        RMIX_LOG((RMIX_WARN(no protocol specified)))
        errno = EINVAL;
        return -1;
      } else {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
        errno = errno2;
        return -1;
      }
    }
  }
  /* Check binding parameter. */
  if (((NULL == binding)||('\0' == binding[0]))&&(NULL != parameters)) {
    /* Search for additional BINDING parameter. */
    if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
        (ESRCH != errno)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in remoteref_create function. */
  if (0 != rmix_providers_symbol2(&remoteref_create, *provider,
                                  "remoteref_create")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_create function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(*provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_create function. */
  if (NULL == remoteref_create) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_create function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(*provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != ((rmix_providers_remoteref_create_t)remoteref_create)
            (remoteref, protocol, binding, parameters, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(*provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference using the appropiate provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create5"
int rmix_remoteref_create5 (rmix_remoteref_t        **remoteref,
                            const rmix_parameters_t  *parameters) {
  const char   *protocol, *binding, *objectidstr;
  void         *remoteref_create;
  unsigned int  objectid, provider;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for protocol parameter. */
  if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no protocol specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Search for object id parameter. */
  if (0 != rmix_parameters_search(parameters, &objectidstr, "OBJECTID")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no object id specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for object id  parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Convert object id parameter to object id. */
  objectid = (unsigned int)atoi(objectidstr);
  /* Search for additional binding parameter. */
  if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
      (ESRCH != errno)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
    errno = errno2;
    return -1;
  }
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in remoteref_create function. */
  if (0 != rmix_providers_symbol2(&remoteref_create, provider,
                                  "remoteref_create")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_create function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_create function. */
  if (NULL == remoteref_create) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_create function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != ((rmix_providers_remoteref_create_t)remoteref_create)
            (remoteref, protocol, binding, parameters, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Unload protocol provider plug-in. */
  if (0 != rmix_providers_unload(provider)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unload protocol provider plug-in)))
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference without loading the provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The preloaded provider plug-in handle.
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create5_p"
int rmix_remoteref_create5_p (rmix_remoteref_t        **remoteref,
                              unsigned int              provider,
                              const rmix_parameters_t  *parameters) {
  const char   *protocol, *binding, *objectidstr;
  void         *remoteref_create;
  unsigned int  objectid;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for protocol parameter. */
  if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no protocol specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Search for object id parameter. */
  if (0 != rmix_parameters_search(parameters, &objectidstr, "OBJECTID")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no object id specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for object id  parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Convert object id parameter to object id. */
  objectid = (unsigned int)atoi(objectidstr);
  /* Search for additional binding parameter. */
  if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
      (ESRCH != errno)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in remoteref_create function. */
  if (0 != rmix_providers_symbol2(&remoteref_create, provider,
                                  "remoteref_create")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_create function)))
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_create function. */
  if (NULL == remoteref_create) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_create function does not exist)))
    errno = ENOTSUP;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != ((rmix_providers_remoteref_create_t)remoteref_create)
            (remoteref, protocol, binding, parameters, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference using the appropiate provider plug-in and
 * without un loading the provider plug-in, while forcing a specific object
 * handle. This is very useful for accessing standard object exports at the
 * client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The provider plug-in handle (return).
 * parameters = The parameters must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create5_p2"
int rmix_remoteref_create5_p2 (rmix_remoteref_t        **remoteref,
                               unsigned int             *provider,
                               const rmix_parameters_t  *parameters) {
  const char   *protocol, *binding, *objectidstr;
  void         *remoteref_create;
  unsigned int  objectid;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check provider parameter. */
  if (NULL == provider) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(provider parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for protocol parameter. */
  if (0 != rmix_parameters_search(parameters, &protocol, "PROTOCOL")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no protocol specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for protocol parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Search for object id parameter. */
  if (0 != rmix_parameters_search(parameters, &objectidstr, "OBJECTID")) {
    if (ESRCH == errno) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(no object id specified)))
      errno = EINVAL;
      return -1;
    } else {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to search for object id  parameter)))
      errno = errno2;
      return -1;
    }
  }
  /* Convert object id parameter to object id. */
  objectid = (unsigned int)atoi(objectidstr);
  /* Search for additional binding parameter. */
  if ((0     != rmix_parameters_search(parameters, &binding, "BINDING"))&&
      (ESRCH != errno)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to search for binding parameter)))
    errno = errno2;
    return -1;
  }
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(provider, protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in remoteref_create function. */
  if (0 != rmix_providers_symbol2(&remoteref_create, *provider,
                                  "remoteref_create")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_create function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(*provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_create function. */
  if (NULL == remoteref_create) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_create function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(*provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != ((rmix_providers_remoteref_create_t)remoteref_create)
            (remoteref, protocol, binding, parameters, objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(*provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference using the appropiate provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create6"
int rmix_remoteref_create6 (rmix_remoteref_t **remoteref,
                            const char        *parameters) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create papameters. */
  if (0 != rmix_parameters_create4(&parameters2, parameters)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create parameters)))
    errno = errno2;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create5(remoteref, parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Destroy papameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy papameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference without loading the provider plug-in,
 * while forcing a specific object id. This is very useful for accessing
 * standard object exports at the client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The preloaded provider plug-in handle.
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create6_p"
int rmix_remoteref_create6_p (rmix_remoteref_t **remoteref,
                              unsigned int       provider,
                              const char        *parameters) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create papameters. */
  if (0 != rmix_parameters_create4(&parameters2, parameters)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create parameters)))
    errno = errno2;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create5_p(remoteref, provider, parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Destroy papameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy papameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Creates a remote object reference using the appropiate provider plug-in and
 * without un loading the provider plug-in, while forcing a specific object
 * handle. This is very useful for accessing standard object exports at the
 * client side.
 *
 * remoteref  = The remote object reference to create (return).
 * provider   = The provider plug-in handle (return).
 * parameters = The parameter pairs must include PROTOCOL (for protocol) and
 *              OBJECTID (for objectid), and may include BINDING and additional
 *              protocol specific parameters (may not be NULL).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_create6_p2"
int rmix_remoteref_create6_p2 (rmix_remoteref_t **remoteref,
                               unsigned int      *provider,
                               const char        *parameters) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check provider parameter. */
  if (NULL == provider) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(provider parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create papameters. */
  if (0 != rmix_parameters_create4(&parameters2, parameters)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create parameters)))
    errno = errno2;
    return -1;
  }
  /* Create remote object reference. */
  if (0 != rmix_remoteref_create5_p2(remoteref, provider, parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create remote object reference)))
    /* Destroy papameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy papameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    /* Destroy remote object reference. */
    if (0 != rmix_remoteref_destroy(remoteref)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
    }
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(*provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Returs a remote object reference in parameters form using the appropiate
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_parameters"
int rmix_remoteref_parameters (rmix_parameters_t      **parameters,
                               const rmix_remoteref_t  *remoteref) {
  void         *remoteref_parameters;
  unsigned int  provider;
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref->protocol parameter. */
  if (NULL == remoteref->protocol) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref->protocol parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Load protocol provider plug-in. */
  if (0 != rmix_providers_load(&provider, remoteref->protocol)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to load protocol provider plug-in)))
    errno = errno2;
    return -1;
  }
  /* Resolve provider plug-in remoteref_parameters function. */
  if (0 != rmix_providers_symbol2(&remoteref_parameters, provider,
                                  "remoteref_parameters")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_parameters function)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_parameters function. */
  if (NULL == remoteref_parameters) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_parameters function does not exist)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = ENOTSUP;
    return -1;
  }
  /* Convert remote object reference. */
  if (0 != ((rmix_providers_remoteref_parameters_t)remoteref_parameters)
            (parameters, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to convert remote object reference)))
    /* Unload protocol provider plug-in. */
    if (0 != rmix_providers_unload(provider)) {
      RMIX_LOG((RMIX_WARN(unable to unoad protocol provider plug-in)))
    }
    errno = errno2;
    return -1;
  }
  /* Unload protocol provider plug-in. */
  if (0 != rmix_providers_unload(provider)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unload protocol provider plug-in)))
    /* Destroy parameters. */
    if (0 != rmix_parameters_destroy(parameters)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Returs a remote object reference in parameters form without loading the
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * provider   = The preloaded provider plug-in handle.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_parameters_p"
int rmix_remoteref_parameters_p (rmix_parameters_t      **parameters,
                                 unsigned int             provider,
                                 const rmix_remoteref_t  *remoteref) {
  void *remoteref_parameters;
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Resolve provider plug-in remoteref_parameters function. */
  if (0 != rmix_providers_symbol2(&remoteref_parameters, provider,
                                  "remoteref_parameters")) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to resolve remoteref_parameters function)))
    errno = errno2;
    return -1;
  }
  /* Check provider plug-in remoteref_parameters function. */
  if (NULL == remoteref_parameters) {
    errno = ENOTSUP;
    RMIX_LOG((RMIX_WARN(remoteref_parameters function does not exist)))
    errno = ENOTSUP;
    return -1;
  }
  /* Convert remote object reference. */
  if (0 != ((rmix_providers_remoteref_parameters_t)remoteref_parameters)
            (parameters, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to convert remote object reference)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Returs a remote object reference in parameters form using the appropiate
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_parameters2"
int rmix_remoteref_parameters2 (char                   **parameters,
                                const rmix_remoteref_t  *remoteref) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Convert remote object reference. */
  if (0 != rmix_remoteref_parameters(&parameters2, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to convert remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Convert parameters. */
  if (0 != rmix_parameters_string(parameters, parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to convert parameters)))
    /* Destroy parameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy parameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    /* Deallocate parameters string. */
    if (NULL != *parameters) {
      free(*parameters);
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Returs a remote object reference in parameters form without loading the
 * provider plug-in.
 *
 * parameters = The parameters form to create.
 * provider   = The preloaded provider plug-in handle.
 * remoteref  = The remote object reference to convert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_parameters2_p"
int rmix_remoteref_parameters2_p (char                   **parameters,
                                  unsigned int             provider,
                                  const rmix_remoteref_t  *remoteref) {
  rmix_parameters_t *parameters2;
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Convert remote object reference. */
  if (0 != rmix_remoteref_parameters_p(&parameters2, provider, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to convert remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Convert parameters. */
  if (0 != rmix_parameters_string(parameters, parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to convert parameters)))
    /* Destroy parameters. */
    if (0 != rmix_parameters_destroy(&parameters2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    }
    errno = errno2;
    return -1;
  }
  /* Destroy parameters. */
  if (0 != rmix_parameters_destroy(&parameters2)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy parameters)))
    /* Deallocate parameters string. */
    if (NULL != *parameters) {
      free(*parameters);
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Duplicates a remote object reference.
 *
 * remoteref1 = The remote object reference to create (return).
 * remoteref2 = The remote object reference to duplicate.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_duplicate"
int rmix_remoteref_duplicate (rmix_remoteref_t       **remoteref1,
                              const rmix_remoteref_t  *remoteref2) {
  unsigned int index;
#ifdef DEBUG
  /* Check remoteref1 parameter. */
  if (NULL == remoteref1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref2 parameter. */
  if (NULL == remoteref2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate remote object reference. */
  if (NULL == (*remoteref1 = malloc(sizeof(rmix_remoteref_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate remote object reference)))
    errno = errno2;
    return -1;
  }
  /* Duplicate protocol. */
  if (NULL == ((*remoteref1)->protocol = strdup(remoteref2->protocol))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate protocol)))
    /* Deallocate remote object reference. */
    free(*remoteref1);
    errno = errno2;
    return -1;
  }
  /* Check binding parameter. */
  if (NULL != remoteref2->binding) {
    /* Duplicate binding. */
    if (NULL == ((*remoteref1)->binding = strdup(remoteref2->binding))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate binding)))
      /* Deallocate protocol duplicate. */
      free((*remoteref1)->protocol);
      /* Deallocate remote object reference. */
      free(*remoteref1);
      errno = errno2;
      return -1;
    }
  } else {
    (*remoteref1)->binding = NULL;
  }
  /* Check for empty interfaces list. */
  if (0 == remoteref2->interfaces.count) {
    /* Initialize interfaces list. */
    (*remoteref1)->interfaces.count = 0;
    (*remoteref1)->interfaces.array = NULL;
  } else {
    /* Allocate interfaces list array. */
    if (NULL == ((*remoteref1)->interfaces.array =
                 malloc((remoteref2->interfaces.count + 1) * sizeof(char*)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to allocate interfaces list array)))
      /* Deallocate binding duplicate. */
      free((*remoteref1)->binding);
      /* Deallocate protocol duplicate. */
      free((*remoteref1)->protocol);
      /* Deallocate remote object reference. */
      free(*remoteref1);
      errno = errno2;
      return -1;
    }
    /* Duplicate interfaces list. */
    for ((*remoteref1)->interfaces.count = 0,
         index = 0; index < remoteref2->interfaces.count; index++) {
      /* Duplicate interfaces list entry. */
      if (NULL == ((*remoteref1)->interfaces.array[index] =
                    strdup(remoteref2->interfaces.array[index]))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to duplicate interfaces list entry)))
        /* Deallocate interfaces list entries. */
        for (index = 0; index < (*remoteref1)->interfaces.count; index++) {
          /* Deallocate interfaces list entry. */
          free((*remoteref1)->interfaces.array[index]);
        }
        /* Deallocate interfaces list array. */
        free((*remoteref1)->interfaces.array);
        /* Deallocate binding duplicate. */
        free((*remoteref1)->binding);
        /* Deallocate protocol duplicate. */
        free((*remoteref1)->protocol);
        /* Deallocate remote object reference. */
        free(*remoteref1);
        errno = errno2;
        return -1;
      }
      (*remoteref1)->interfaces.count++;
    }
    /* Terminate interfaces list. */
    (*remoteref1)->interfaces.array[index] = NULL;
  }
  /* Duplicate remote object reference location. */
  if (NULL == ((*remoteref1)->location = strdup(remoteref2->location))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference location)))
    /* Deallocate interfaces list entries. */
    for (index = 0; index < (*remoteref1)->interfaces.count; index++) {
      /* Deallocate interfaces list entry. */
      free((*remoteref1)->interfaces.array[index]);
    }
    /* Deallocate interfaces list array. */
    if (NULL != (*remoteref1)->interfaces.array) {
      free((*remoteref1)->interfaces.array);
    }
    /* Deallocate binding duplicate. */
    free((*remoteref1)->binding);
    /* Deallocate protocol duplicate. */
    free((*remoteref1)->protocol);
    /* Deallocate remote object reference. */
    free(*remoteref1);
    errno = errno2;
    return -1;
  }
  /* Duplicate object identifier. */
  if (NULL == ((*remoteref1)->object = strdup(remoteref2->object))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object identifier)))
    /* Deallocate remote object reference location duplicate. */
    free((*remoteref1)->location);
    /* Deallocate interfaces list entries. */
    for (index = 0; index < (*remoteref1)->interfaces.count; index++) {
      /* Deallocate interfaces list entry. */
      free((*remoteref1)->interfaces.array[index]);
    }
    /* Deallocate interfaces list array. */
    if (NULL != (*remoteref1)->interfaces.array) {
      free((*remoteref1)->interfaces.array);
    }
    /* Deallocate binding duplicate. */
    free((*remoteref1)->binding);
    /* Deallocate protocol duplicate. */
    free((*remoteref1)->protocol);
    /* Deallocate remote object reference. */
    free(*remoteref1);
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Destroys a remote object reference.
 *
 * remoteref = The remote object reference to destroy (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_destroy"
int rmix_remoteref_destroy (rmix_remoteref_t **remoteref) {
  unsigned int index;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check *remoteref parameter. */
  if (NULL == *remoteref) {
    return 0;
  }
  /* Deallocate object identifier duplicate. */
  free((*remoteref)->object);
  /* Deallocate remote object reference location duplicate. */
  free((*remoteref)->location);
  /* Deallocate interfaces list entries. */
  for (index = 0; index < (*remoteref)->interfaces.count; index++) {
    /* Deallocate interfaces list entry. */
    free((*remoteref)->interfaces.array[index]);
  }
  /* Deallocate interfaces list array. */
  if (NULL != (*remoteref)->interfaces.array) {
    free((*remoteref)->interfaces.array);
  }
  /* Deallocate binding duplicate. */
  if (NULL != (*remoteref)->binding) {
    free((*remoteref)->binding);
  }
  /* Deallocate protocol duplicate. */
  free((*remoteref)->protocol);
  /* Deallocate remote object reference. */
  free(*remoteref);
  *remoteref = NULL;
  return 0;
}

/*
 * Gets the object id from a remote object reference.
 *
 * remoteref = The remote object reference.
 * objectid  = The object id (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_objectid"
int rmix_remoteref_objectid (const rmix_remoteref_t *remoteref,
                             unsigned int           *objectid) {
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref->object parameter. */
  if (NULL == remoteref->object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref->object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check objectid parameter. */
  if (NULL == objectid) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(objectid parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Reconstruct object id. */
  if (1 != sscanf(remoteref->object, "%*u:%*u:%u:%*u", objectid)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reconstruct object id)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Compares two remote object references.
 *
 * diff       = The difference between both remote object references (return).
 *              diff > 0 if remoteref1 < remoteref2,
 *              diff < 0 if remoteref1 > remoteref2, and
 *              diff = 0 if remoteref1 = remoteref2
 * remoteref1 = The remote object reference to compare against.
 * remoteref2 = The remote object reference to compare with.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_remoteref_compare"
int rmix_remoteref_compare (int                    *diff,
                            const rmix_remoteref_t *remoteref1,
                            const rmix_remoteref_t *remoteref2) {
  unsigned int index;
#ifdef DEBUG
  /* Check diff parameter. */
  if (NULL == diff) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(diff parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref1 parameter. */
  if (NULL == remoteref1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref2 parameter. */
  if (NULL == remoteref2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Compare and check for equal protocols,
     bindings, locations and object identifiers. */
  if ((0 == (*diff = strcmp(remoteref1->protocol, remoteref2->protocol  )))&&
      ((( NULL == remoteref1->binding) && (NULL == remoteref2->binding ) )||
       (((NULL != remoteref1->binding) && (NULL != remoteref2->binding ))&&
        (0 == (*diff = strcmp(remoteref1->binding, remoteref2->binding)))))&&
      (0 == (*diff = strcmp(remoteref1->location, remoteref2->location  )))&&
      (0 == (*diff = strcmp(remoteref1->object,   remoteref2->object    )))&&
      (0 == (*diff = remoteref2->interfaces.count -
                     remoteref1->interfaces.count))) {
    /* Compare and check for equal remote object reference interfaces. */
    for (index = 0; index < remoteref1->interfaces.count; index++) {
      /* Compare and check for equal remote object reference interface. */
      if (0 != (*diff = strcmp(remoteref1->interfaces.array[index],
                               remoteref2->interfaces.array[index]))) {
        return 0; 
      }
    }
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
