/******************************************************************************
 *
 * Source file for rmixregistry (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <rmixregistry.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Server-side method descriptors for rmix registry. */
const rmix_method_t rmixregistry_methods[RMIXREGISTRY_METHODS_COUNT] =
  RMIXREGISTRY_METHODS;

/* Server-side interface for rmix registry. */
const rmix_interface_t rmixregistry_interface = {
  RMIXREGISTRY_METHODS_COUNT,           /* method descriptor count */
  (rmix_method_t*)rmixregistry_methods  /* method descriptor array */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Creates a RMIX registry.
 *
 * rmixregistry = The RMIX registry to create.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_create"
int rmixregistry_create (rmixregistry_t **rmixregistry) {
  int error;
  /* Allocate rmixregistry. */
  if (NULL == (*rmixregistry = malloc(sizeof(rmixregistry_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate rmixregistry)))
    errno = errno2;
    return -1;
  }
  /* Initialize rmixregistry mutex. */
  if (0 != (error = pthread_mutex_init(&(*rmixregistry)->mutex, NULL))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to initialize rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    /* Deallocate rmixregistry. */
    free(*rmixregistry);
    *rmixregistry = NULL;
    errno = error;
    return -1;
  }
  /* Initialize rmixregistry registry. */
  (*rmixregistry)->count    = 0;
  (*rmixregistry)->registry = NULL;
  return 0;
}

/*
 * Destroys a RMIX registry.
 *
 * rmixregistry = The RMIX registry to destroy.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_destroy"
int rmixregistry_destroy (rmixregistry_t **rmixregistry) {
  int          error, result, errno2;
  unsigned int index;
  result = 0;
  errno2 = 0;
  /* Lock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_lock(&(*rmixregistry)->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Destroy registry entries. */
  for (index = 0; index < (*rmixregistry)->count; index++) {
    /* Deallocate name. */
    free((*rmixregistry)->registry[index].name);
    /* Destroy bound remote object reference. */
    if (0 !=
        rmix_remoteref_destroy(&(*rmixregistry)->registry[index].remoteref)) {
      errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference)))
      result = -1;
    }
  }
  /* Reset registry size. */
  (*rmixregistry)->count = 0;
  /* Deallocate registry. */
  if (NULL != (*rmixregistry)->registry) {
    free((*rmixregistry)->registry);
    (*rmixregistry)->registry = 0;
  }
  /* Unlock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_unlock(&(*rmixregistry)->mutex))) {
    errno2 = error;
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    result = -1;
  }
  /* Finalize rmixregistry mutex. */
  if (0 != (error = pthread_mutex_destroy(&(*rmixregistry)->mutex))) {
    errno2 = error;
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to finalize rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    result = -1;
  }
  /* Deallocate rmixregistry. */
  free(*rmixregistry);
  *rmixregistry = NULL;
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Exports a RMIX registry without forcing a specific object handle.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported registry (return).
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export"
int rmixregistry_export (rmixregistry_t           *rmixregistry,
                         rmix_localref_t         **localref,
                         const char               *protocol,
                         const char               *binding,
                         const rmix_parameters_t  *parameters) {
  /* Export the registry. */
  if (0 != rmix_export(localref, protocol, binding, parameters, rmixregistry,
                       &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a RMIX registry without forcing a specific object handle and without
 * loading the provider plug-in.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported registry (return).
 * provider     = The preloaded provider plug-in handle.
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export_p"
int rmixregistry_export_p (rmixregistry_t           *rmixregistry,
                           rmix_localref_t         **localref,
                           unsigned int              provider,
                           const char               *protocol,
                           const char               *binding,
                           const rmix_parameters_t  *parameters) {
  /* Export the registry. */
  if (0 != rmix_export_p(localref, provider, protocol, binding, parameters,
                         rmixregistry, &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a RMIX registry while forcing a specific object handle.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported registry (return).
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 * handle       = The object handle.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export2"
int rmixregistry_export2 (rmixregistry_t           *rmixregistry,
                          rmix_localref_t         **localref,
                          const char               *protocol,
                          const char               *binding,
                          const rmix_parameters_t  *parameters,
                          unsigned int              handle) {
  /* Export the registry. */
  if (0 != rmix_export2(localref, protocol, binding, parameters, handle,
                        rmixregistry, &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a RMIX registry while forcing a specific object handle and without
 * loading the provider plug-in.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported registry (return).
 * provider     = The preloaded provider plug-in handle.
 * protocol     = The communication protocol. Overrides any additional PROTOCOL
 *                parameter. May only be NULL or "" if additional parameters
 *                contains PROTOCOL.
 * binding      = The binding  (protocol specific) (may be NULL, ""). Overrides
 *                any additional BINDING parameter.
 * parameters   = Additional parameters (protocol specific) (may be NULL, "").
 * handle       = The object handle.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export2_p"
int rmixregistry_export2_p (rmixregistry_t           *rmixregistry,
                            rmix_localref_t         **localref,
                            unsigned int              provider,
                            const char               *protocol,
                            const char               *binding,
                            const rmix_parameters_t  *parameters,
                            unsigned int              handle) {
  /* Export the registry. */
  if (0 != rmix_export2_p(localref, provider, protocol, binding, parameters,
                          handle, rmixregistry,
                          &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a RMIX registry while forcing a specific object handle. This is very
 * useful for standard object exports.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported object (return).
 * parameters   = The parameters must include PROTOCOL (for protocol) and
 *                OBJECTID (for handle), and may include BINDING and additional
 *                protocol specific parameters (may not be NULL).
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export3"
int rmixregistry_export3 (rmixregistry_t           *rmixregistry,
                          rmix_localref_t         **localref,
                          const rmix_parameters_t  *parameters) {
  /* Export the registry. */
  if (0 != rmix_export3(localref, parameters,
                        rmixregistry, &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a RMIX registry while forcing a specific object handle and without
 * loading the provider plug-in. This is very useful for standard object
 * exports.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported object (return).
 * provider     = The preloaded provider plug-in handle.
 * parameters   = The parameters must include PROTOCOL (for protocol) and
 *                OBJECTID (for handle), and may include BINDING and additional
 *                protocol specific parameters (may not be NULL).
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export3_p"
int rmixregistry_export3_p (rmixregistry_t           *rmixregistry,
                            rmix_localref_t         **localref,
                            unsigned int              provider,
                            const rmix_parameters_t  *parameters) {
  /* Export the registry. */
  if (0 != rmix_export3_p(localref, provider, parameters,
                          rmixregistry, &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a RMIX registry while forcing a specific object handle. This is very
 * useful for standard object exports.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported object (return).
 * parameters   = The parameter pairs must include PROTOCOL (for protocol) and
 *                OBJECTID (for handle), and may include BINDING and additional
 *                protocol specific parameters (may not be NULL).
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export4"
int rmixregistry_export4 (rmixregistry_t   *rmixregistry,
                          rmix_localref_t **localref,
                          const char       *parameters) {
  /* Export the registry. */
  if (0 != rmix_export4(localref, parameters,
                        rmixregistry, &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Exports a RMIX registry while forcing a specific object handle and without
 * loading the provider plug-in. This is very useful for standard object
 * exports.
 *
 * rmixregistry = The RMIX registry to export.
 * localref     = The local object reference of the exported object (return).
 * provider     = The preloaded provider plug-in handle.
 * parameters   = The parameter pairs must include PROTOCOL (for protocol) and
 *                OBJECTID (for handle), and may include BINDING and additional
 *                protocol specific parameters (may not be NULL).
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_export4_p"
int rmixregistry_export4_p (rmixregistry_t   *rmixregistry,
                            rmix_localref_t **localref,
                            unsigned int      provider,
                            const char       *parameters) {
  /* Export the registry. */
  if (0 != rmix_export4_p(localref, provider, parameters,
                          rmixregistry, &rmixregistry_interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to export the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Unexports the registry.
 *
 * localref = The local object reference of the exported object (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_unexport"
int rmixregistry_unexport (rmix_localref_t **localref) {
  /* Unexport the registry. */
  if (0 != rmix_unexport(localref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unexport the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Unexports a RMIX registry without unloading the provider plug-in.
 *
 * localref = The local object reference of the exported registry (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_unexport"
int rmixregistry_unexport_p (rmix_localref_t **localref) {
  /* Unexport the registry. */
  if (0 != rmix_unexport_p(localref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to unexport the registry)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Binds a name with a remote object reference at a RMIX registry.
 *
 * rmixregistry = The RMIX registry.
 * name         = The name to bind.
 * remoteref    = The remote object reference to bind.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_bind"
int rmixregistry_bind (rmixregistry_t         *rmixregistry,
                       const char             *name,
                       const rmix_remoteref_t *remoteref) {
  int               error;
  char             *name2;
  void             *registry;
  unsigned int      index;
  rmix_remoteref_t *remoteref2;
#ifdef DEBUG
  /* Check rmixregistry parameter. */
  if (NULL == rmixregistry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(rmixregistry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Duplicate name. */
  if (NULL == (name2 = strdup(name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate name)))
    errno = errno2;
    return -1;
  }
  /* Duplicate remote object reference. */
  if (0 != rmix_remoteref_duplicate(&remoteref2, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference)))
    /* Deallocate name duplicate. */
    free(name2);
    errno = errno2;
    return -1;
  }
  /* Lock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy remote object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&remoteref2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    }
    /* Deallocate name duplicate. */
    free(name2);
    errno = error;
    return -1;
  }
  /* Search for name in registry. */
  for (index = 0; index < rmixregistry->count; index++) {
    if (0 == strcmp(rmixregistry->registry[index].name, name)) {
      break;
    }
  }
  /* Check if name is in registry. */
  if (index < rmixregistry->count) {
    errno = EALREADY;
    RMIX_LOG((RMIX_WARN(name is already in registry)))
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy remote object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&remoteref2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    }
    /* Deallocate name duplicate. */
    free(name2);
    errno = EALREADY;
    return -1;
  }
  /* Increase registry size. */
  index = rmixregistry->count;
  rmixregistry->count++;
  /* Reallocate registry. */
  if (NULL == (registry = realloc(rmixregistry->registry,
                                  rmixregistry->count *
                           sizeof(rmixregistry->registry[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate registry)))
    rmixregistry->count--;
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy remote object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&remoteref2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    }
    /* Deallocate name duplicate. */
    free(name2);
    errno = errno2;
    return -1;
  }
  rmixregistry->registry = registry;
  /* Save registry entry. */
  rmixregistry->registry[index].name      = name2;
  rmixregistry->registry[index].remoteref = remoteref2;
  /* Unlock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    return -1;
  }
  return 0;
}

/*
 * Returns an array of the names in a RMIX registry.
 *
 * rmixregistry = The RMIX registry.
 * count        = The number of names in the registry (return).
 * names        = The NULL terminated array of names (return).
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_list"
int rmixregistry_list (rmixregistry_t   *rmixregistry,
                       unsigned int     *count,
                       char           ***names) {
  int          error;
  unsigned int index;
#ifdef DEBUG
  /* Check rmixregistry parameter. */
  if (NULL == rmixregistry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(rmixregistry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if (NULL == count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check names parameter. */
  if (NULL == names) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(names parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Allocate array of names. */
  if (NULL == (*names = malloc((rmixregistry->count + 1) * sizeof(char*)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate array of names)))
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate names in registry. */
  for (index = 0, (*count) = 0;
       index < rmixregistry->count; index++, (*count)++) {
    /* Duplicate name. */
    if (NULL == ((*names)[index] =
                 strdup(rmixregistry->registry[index].name))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate name)))
      /* Deallocate name duplicates. */
      for (index = 0; index < (*count); index++) {
        free((*names)[index]);
      }
      /* Deallocate array of names. */
      free(*names);
      /* Unlock rmixregistry mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = errno2;
      return -1;
    }
  }
  /* Terminate array of names. */
  (*names)[index] = NULL;
  /* Unlock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Returns the remote object reference associated with a name at a RMIX
 * registry. 
 *
 * rmixregistry = The RMIX registry.
 * name         = The name to lookup.
 * remoteref    = The remote object reference (return).
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_lookup"
int rmixregistry_lookup (rmixregistry_t    *rmixregistry,
                         const char        *name,
                         rmix_remoteref_t **remoteref) {
  int          error;
  unsigned int index;
#ifdef DEBUG
  /* Check rmixregistry parameter. */
  if (NULL == rmixregistry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(rmixregistry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Search for name in registry. */
  for (index = 0; index < rmixregistry->count; index++) {
    if (0 == strcmp(rmixregistry->registry[index].name, name)) {
      break;
    }
  }
  /* Check if name is not in registry. */
  if (index == rmixregistry->count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(name is not in registry)))
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Duplicate remote object reference. */
  if (0 != rmix_remoteref_duplicate(remoteref, 
             rmixregistry->registry[index].remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference)))
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Unlock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at a RMIX registry.
 *
 * rmixregistry = The RMIX registry.
 * name         = The name to bind.
 * remoteref    = The remote object reference to bind.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_rebind"
int rmixregistry_rebind (rmixregistry_t         *rmixregistry,
                         const char             *name,
                         const rmix_remoteref_t *remoteref) {
  int                error;
  char              *name2;
  void              *registry;
  unsigned int       index;
  rmix_remoteref_t  *remoteref2, *remoteref3;
#ifdef DEBUG
  /* Check rmixregistry parameter. */
  if (NULL == rmixregistry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(rmixregistry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Duplicate name. */
  if (NULL == (name2 = strdup(name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate name)))
    errno = errno2;
    return -1;
  }
  /* Duplicate remote object reference. */
  if (0 != rmix_remoteref_duplicate(&remoteref2, remoteref)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate remote object reference)))
    /* Deallocate name duplicate. */
    free(name2);
    errno = errno2;
    return -1;
  }
  /* Lock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy remote object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&remoteref2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    }
    /* Deallocate name duplicate. */
    free(name2);
    errno = error;
    return -1;
  }
  /* Search for name in registry. */
  for (index = 0; index < rmixregistry->count; index++) {
    if (0 == strcmp(rmixregistry->registry[index].name, name)) {
      break;
    }
  }
  /* Check if name is in registry. */
  if (index < rmixregistry->count) {
    /* Deallocate name duplicate. */
    free(name2);
    /* Save old remote object reference duplicate from registry entry. */
    remoteref3 = rmixregistry->registry[index].remoteref;
    /* Replace remote object reference duplicate in registry entry. */
    rmixregistry->registry[index].remoteref = remoteref2;
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      errno = error;
      rmix_syserr();
      /* Destroy old remote object reference duplicate. */
      if (0 != rmix_remoteref_destroy(&remoteref2)) {
        RMIX_LOG((
          RMIX_WARN(unable to destroy old remote object reference duplicate)))
      }
      errno = error;
      return -1;
    }
    /* Destroy old remote object reference duplicate. */
    if (0 != rmix_remoteref_destroy(&remoteref2)) {
      int errno2 = errno;
      RMIX_LOG((
        RMIX_WARN(unable to destroy old remote object reference duplicate)))
      errno = errno2;
      rmix_syserr();
      errno = errno2;
      return -1;
    }
  } else {
    /* Increase registry size. */
    index = rmixregistry->count;
    rmixregistry->count++;
    /* Reallocate registry. */
    if (NULL == (registry = realloc(rmixregistry->registry,
                                    rmixregistry->count *
                             sizeof(rmixregistry->registry[0])))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to reallocate registry)))
      rmixregistry->count--;
      /* Unlock rmixregistry mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
        errno = errno2;
        rmix_syserr();
      }
      /* Destroy remote object reference duplicate. */
      if (0 != rmix_remoteref_destroy(&remoteref2)) {
        RMIX_LOG((
          RMIX_WARN(unable to destroy remote object reference duplicate)))
      }
      /* Deallocate name duplicate. */
      free(name2);
      errno = errno2;
      return -1;
    }
    rmixregistry->registry = registry;
    /* Save registry entry. */
    rmixregistry->registry[index].name      = name2;
    rmixregistry->registry[index].remoteref = remoteref2;
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      errno = error;
      rmix_syserr();
      errno = error;
      return -1;
    }
  }
  return 0;
}

/*
 * Unbinds a name from a remote object reference at a RMIX registry.
 *
 * rmixregistry = The RMIX registry.
 * name         = The name to unbind.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_unbind"
int rmixregistry_unbind (rmixregistry_t *rmixregistry,
                         const char     *name) {
  int           error, result, errno2;
  void         *registry;
  unsigned int  index;
#ifdef DEBUG
  /* Check rmixregistry parameter. */
  if (NULL == rmixregistry) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(rmixregistry parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    return -1;
  }
  /* Search for name in registry. */
  for (index = 0; index < rmixregistry->count; index++) {
    if (0 == strcmp(rmixregistry->registry[index].name, name)) {
      break;
    }
  }
  /* Check if name is not in registry. */
  if (index == rmixregistry->count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(name is not in registry)))
    /* Unlock rmixregistry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
      errno = error;
      rmix_syserr();
    }
    return -1;
  }
  result = 0;
  errno2 = 0;
  /* Destroy remote object reference duplicate. */
  if (0 != rmix_remoteref_destroy(
             &rmixregistry->registry[index].remoteref)) {
    errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to destroy remote object reference duplicate)))
    result = -1;
  }
  /* Deallocate name duplicate. */
  free(rmixregistry->registry[index].name);
  /* Remove entry from registry. */
  rmixregistry->count--;
  memmove(rmixregistry->registry + index,
          rmixregistry->registry + index + 1,
         (rmixregistry->count - index) *
   sizeof(rmixregistry->registry[0]));
  /* Reallocate registry. */
  if (0 == rmixregistry->count) {
    free(rmixregistry->registry);
    rmixregistry->registry = NULL;
  } else if (NULL == (registry = realloc(rmixregistry->registry,
                                         rmixregistry->count *
                                  sizeof(rmixregistry->registry[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate registry)))
    errno = 0;
  } else {
    rmixregistry->registry = registry;
  }
  /* Unlock rmixregistry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmixregistry->mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock rmixregistry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Invokes rmixregistry_bind (server-side method stub).
 * Binds a name with a remote object reference at an exported RMIX registry.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_bind_call"
RMIX_METHOD_CALL(rmixregistry_bind_call) {
  int               result;
  const char       *name;
  rmix_remoteref_t  remoteref;
#ifdef DEBUG
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name                       =                           inary[0]     ;
  remoteref.protocol         =  (char*        )(((void**)inary[1])[0]);
  remoteref.binding          =  (char*        )(((void**)inary[1])[1]);
  remoteref.interfaces.count = *(unsigned int*)(((void**)inary[1])[2]);
  remoteref.interfaces.array =  (char**       )(((void**)inary[1])[3]);
  remoteref.location         =  (char*        )(((void**)inary[1])[4]);
  remoteref.object           =  (char*        )(((void**)inary[1])[5]);
  /* Call method function. */
  result = rmixregistry_bind(object, name, &remoteref);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes rmixregistry_list (server-side method stub).
 * Returns an array of the names in an exported RMIX registry.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_list_call"
RMIX_METHOD_CALL(rmixregistry_list_call) {
  int            result;
  char         **names;
  unsigned int   count;
#ifdef DEBUG
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Call method function. */
  result = rmixregistry_list(object, &count, &names);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Check result. */
  if (0 == result) {
    *(unsigned int*)(outary[1]) = count;
                     outary[2]  = names;
  } else {
    *(unsigned int*)(outary[1]) = 0;
                     outary[2]  = NULL;
  }
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes rmixregistry_lookup (server-side method stub).
 * Returns the remote object reference associated with a name  at an exported
 * RMIX registry. 
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_lookup_call"
RMIX_METHOD_CALL(rmixregistry_lookup_call) {
  int               result;
  const char       *name;
  rmix_remoteref_t *remoteref;
#ifdef DEBUG
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if (NULL == outary) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name      = inary[0];
  remoteref = NULL;
  /* Call method function. */
  result = rmixregistry_lookup(object, name, &remoteref);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Check result. */
  if (0 == result) {
                     ((void**)outary[1])[0]  = remoteref->protocol;
                     ((void**)outary[1])[1]  = remoteref->binding;
    *(unsigned int*)(((void**)outary[1])[2]) = remoteref->interfaces.count;
                     ((void**)outary[1])[3]  = remoteref->interfaces.array;
                     ((void**)outary[1])[4]  = remoteref->location;
                     ((void**)outary[1])[5]  = remoteref->object;
  } else {
                     ((void**)outary[1])[0]  = NULL;
                     ((void**)outary[1])[1]  = NULL;
    *(unsigned int*)(((void**)outary[1])[2]) = 0;
                     ((void**)outary[1])[3]  = NULL;
                     ((void**)outary[1])[4]  = NULL;
                     ((void**)outary[1])[5]  = NULL;
  }
  /* Cleanup. */
  if (NULL != remoteref) {
    free(remoteref);
  }
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes rmixregistry_rebind (server-side method stub).
 * Rebinds (creates a new or replaces an existing binding) a name with a remote
 * object reference at an exported RMIX registry.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_rebind_call"
RMIX_METHOD_CALL(rmixregistry_rebind_call) {
  int               result;
  const char       *name;
  rmix_remoteref_t  remoteref;
#ifdef DEBUG
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name                       =                           inary[0]     ;
  remoteref.protocol         =  (char*        )(((void**)inary[1])[0]);
  remoteref.binding          =  (char*        )(((void**)inary[1])[1]);
  remoteref.interfaces.count = *(unsigned int*)(((void**)inary[1])[2]);
  remoteref.interfaces.array =  (char**       )(((void**)inary[1])[3]);
  remoteref.location         =  (char*        )(((void**)inary[1])[4]);
  remoteref.object           =  (char*        )(((void**)inary[1])[5]);
  /* Call method function. */
  result = rmixregistry_rebind(object, name, &remoteref);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/*
 * Invokes rmixregistry_unbind (server-side method stub).
 * Unbinds a name from a remote object reference at an exported RMIX registry.
 *
 * object = The local object.
 * outary = The output values array. The values array and its containing
 *          values are allocated before the call with the exception of
 *          variable arrays and strings. They are allocated dynamically
 *          or explicitly set to NULL by this function. In the case of
 *          variable arrays, the length value is allocated before the call
 *          and is set to 0 if its variable array is NULL.
 * outcnt = The output values count.
 * inary  = The input values array. The values array and its containing
 *          values are not modified or deallocated by this call.
 * incnt  = The input values count.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_unbind_call"
RMIX_METHOD_CALL(rmixregistry_unbind_call) {
  int         result;
  const char *name;
#ifdef DEBUG
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outary parameter. */
  if ((NULL == outary)&&(0 != outcnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check outcnt parameter. */
  if ((0 == outcnt)&&(NULL != outary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(outcnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check inary parameter. */
  if ((NULL == inary)&&(0 != incnt)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(inary parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check incnt parameter. */
  if ((0 == incnt)&&(NULL != inary)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(incnt parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Prepare input. */
  name = inary[0];
  /* Call method function. */
  result = rmixregistry_unbind(object, name);
  /* Prepare output. */
  *(int*)(outary[0]) = result;
  /* Reset errno if needed. */
  if (0 != result) {
    errno = 0;
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
