/******************************************************************************
 *
 * Source file for rmixregistryd (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <rmixregistryd.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 *
 * Parses the command line options and runs the rmix registry as a daemon.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = EXIT_SUCCESS on success or EXIT_FAILURE on any error.
 */
int main (int    argc,
          char **argv,
          char **envp);

/*
 * The signal handler.
 *
 * signum = The number of the received signal.
 */
void rmixregistry_handler (int signum);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Parses the command line options and runs the rmix registry as a daemon.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = EXIT_SUCCESS on success or EXIT_FAILURE on any error.
 */
#undef  __FUNC__
#define __FUNC__ "main"
int main (int    argc,
          char **argv,
          char **envp) {
  int                       result, signum, error;
  pid_t                     pid;
  sigset_t                  sigmask;
  unsigned int              daemon, pool_minimum, pool_maximum, pool_timeout,
                            jobs_maximum, jobs_timeout, index, count,
                            pool_maximum_old;
  rmixregistry_t           *rmixregistry;  
  rmix_localref_t         **localrefs;
  const rmix_parameters_t **parameters;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv[0] parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Check envp parameter. */
  if (NULL == envp) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envp parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
#endif /* DEBUG */
  /* Initialize options. */
  if (0 != rmixregistryd_options_init(argc, (const char**)argv,
                                            (const char**)envp)) {
    RMIX_LOG((RMIX_WARN(unable to initialize options)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Install signal handler to avoid immediate termination on SIGTERM. */
  if (SIG_ERR == signal(SIGTERM, rmixregistry_handler)) {
    RMIX_LOG((RMIX_WARN(unable to install signal handler)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Initialize signal mask. */
  if (-1 == sigfillset(&sigmask)) {
    RMIX_LOG((RMIX_WARN(unable to fill signal mask)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Block all signals. */
  if (0 != (error = pthread_sigmask(SIG_BLOCK, &sigmask, NULL))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to set thread signal mask)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Reinitialize signal mask. */
  if (-1 == sigemptyset(&sigmask)) {
    RMIX_LOG((RMIX_WARN(unable to fill signal mask)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Add SIGTERM to signal mask. */
  if (-1 == sigaddset(&sigmask, SIGTERM)) {
    RMIX_LOG((RMIX_WARN(unable to add signal to signal mask)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the daemon flag option. */
  if (0 != rmixregistryd_options_daemon_get(&daemon)) {
    RMIX_LOG((RMIX_WARN(unable to get the daemon flag option)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the minimum thread count option. */
  if (0 != rmixregistryd_options_pool_minimum_get(&pool_minimum)) {
    RMIX_LOG((RMIX_WARN(unable to get the minimum thread count option)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the maximum thread count option. */
  if (0 != rmixregistryd_options_pool_maximum_get(&pool_maximum)) {
    RMIX_LOG((RMIX_WARN(unable to get the maximum thread count option)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the exiting thread timeout option. */
  if (0 != rmixregistryd_options_pool_timeout_get(&pool_timeout)) {
    RMIX_LOG((RMIX_WARN(unable to get the exiting thread timeout option)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the the maximum job count option. */
  if (0 != rmixregistryd_options_jobs_maximum_get(&jobs_maximum)) {
    RMIX_LOG((RMIX_WARN(unable to get the the maximum job count option)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the waiting thread timeout option. */
  if (0 != rmixregistryd_options_jobs_timeout_get(&jobs_timeout)) {
    RMIX_LOG((RMIX_WARN(unable to get the waiting thread timeout option)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the maximum thread count. */
  if (0 != rmix_threads_pool_maximum_get(&pool_maximum_old)) {
    RMIX_LOG((RMIX_WARN(unable to get the maximum thread count)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  if (pool_minimum > pool_maximum_old) {
    /* Set the maximum thread count. */
    if (0 != rmix_threads_pool_maximum_set(pool_maximum)) {
      RMIX_LOG((RMIX_WARN(unable to set the maximum thread count)))
      fprintf(stdout, "ERROR: An internal error has occurred.\n");
      /* Finalize options. */
      if (0 != rmixregistryd_options_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize options)))
      }
      return EXIT_FAILURE;
    }
    /* Set the minimum thread count. */
    if (0 != rmix_threads_pool_minimum_set(pool_minimum)) {
      RMIX_LOG((RMIX_WARN(unable to set the minimum thread count)))
      fprintf(stdout, "ERROR: An internal error has occurred.\n");
      return EXIT_FAILURE;
    }
  } else {
    /* Set the minimum thread count. */
    if (0 != rmix_threads_pool_minimum_set(pool_minimum)) {
      RMIX_LOG((RMIX_WARN(unable to set the minimum thread count)))
      fprintf(stdout, "ERROR: An internal error has occurred.\n");
      return EXIT_FAILURE;
    }
    /* Set the maximum thread count. */
    if (0 != rmix_threads_pool_maximum_set(pool_maximum)) {
      RMIX_LOG((RMIX_WARN(unable to set the maximum thread count)))
      fprintf(stdout, "ERROR: An internal error has occurred.\n");
      /* Finalize options. */
      if (0 != rmixregistryd_options_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize options)))
      }
      return EXIT_FAILURE;
    }
  }
  /* Set the exiting thread timeout. */
  if (0 != rmix_threads_pool_timeout_set(pool_timeout)) {
    RMIX_LOG((RMIX_WARN(unable to set the exiting thread timeout)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Set the the maximum job count. */
  if (0 != rmix_threads_jobs_maximum_set(jobs_maximum)) {
    RMIX_LOG((RMIX_WARN(unable to set the the maximum job count)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Set the waiting thread timeout. */
  if (0 != rmix_threads_jobs_timeout_set(jobs_timeout)) {
    RMIX_LOG((RMIX_WARN(unable to set the waiting thread timeout)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Get the export parameters. */
  if (0 != rmixregistryd_options_parameters_get(&parameters)) {
    RMIX_LOG((RMIX_WARN(unable to get the export parameters)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Check the export parameters. */
  if ((NULL == parameters)||(NULL == parameters[0])) {
    RMIX_LOG((RMIX_WARN(missing export parameters)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Check daemon flag option. */
  if (1 == daemon) {
    /* Create daemon process. */
    pid = fork();
  } else {
    /* Do not create daemon process. */
    pid = 0;
  }
  switch (pid) {
    /* Check for error. */
    case -1: {
      RMIX_LOG((RMIX_WARN(unable to create daemon process)))
      fprintf(stdout, "ERROR: An internal error has occurred.\n");
      /* Finalize options. */
      if (0 != rmixregistryd_options_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize options)))
      }
      return EXIT_FAILURE;
    }
    /* Check for child process or if no daemon is created. */
    case 0: {
      /* Allocate local registry references. */
      for (count = 0; NULL != parameters[count]; count++);
      if (NULL == (localrefs = malloc(count * sizeof(rmix_localref_t*)))) {
        RMIX_LOG((RMIX_WARN(unable to allocate local registry references)))
        fprintf(stdout, "ERROR: An internal error has occurred.\n");
        /* Finalize options. */
        if (0 != rmixregistryd_options_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize options)))
        }
        return EXIT_FAILURE;
      }
      /* Initialize rmix. */
      if (0 != rmix_init2()) {
        RMIX_LOG((RMIX_WARN(unable to initialize rmix)))
        fprintf(stdout, "ERROR: An internal error has occurred.\n");
        /* Deallocate local registry references. */
        free(localrefs);
        /* Finalize options. */
        if (0 != rmixregistryd_options_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize options)))
        }
        return EXIT_FAILURE;
      }
      /* Create rmixregistry. */
      if (0 != rmixregistry_create(&rmixregistry)) {
        RMIX_LOG((RMIX_WARN(unable to create rmixregistry)))
        fprintf(stdout, "ERROR: An internal error has occurred.\n");
        /* Finalize rmix. */
        if (0 != rmix_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
        }
        /* Deallocate local registry references. */
        free(localrefs);
        /* Finalize options. */
        if (0 != rmixregistryd_options_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize options)))
        }
        return EXIT_FAILURE;
      }
      /* Export rmixregistry. */
      for (index = 0; index < count; index++) {
        /* Export rmixregistry. */
        if (0 != rmixregistry_export3(rmixregistry, &localrefs[index],
                                      parameters[index])) {
          RMIX_LOG((RMIX_WARN(unable to export %u), index))
          fprintf(stdout, "ERROR: Unable to export # %u.\n", index);
          /* Unexport rmixregistry. */
          while (0 != index--) {
            /* Unexport rmixregistry. */
            if (0 != rmixregistry_unexport(&localrefs[index])) {
              RMIX_LOG((RMIX_WARN(unable to unexport # %u), index))
            }
          }
          /* Destroy rmixregistry. */
          if (0 != rmixregistry_destroy(&rmixregistry)) {
            RMIX_LOG((RMIX_WARN(unable to destroy rmixregistry)))
          }
          /* Finalize rmix. */
          if (0 != rmix_fini()) {
            RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
          }
          /* Deallocate local registry references. */
          free(localrefs);
          /* Finalize options. */
          if (0 != rmixregistryd_options_fini()) {
            RMIX_LOG((RMIX_WARN(unable to finalize options)))
          }
          return EXIT_FAILURE;
        }
      }
      /* Unblock SIGTERM signal. */
      if (0 != (error = pthread_sigmask(SIG_UNBLOCK, &sigmask, NULL))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to set thread signal mask)))
        fprintf(stdout, "ERROR: An internal error has occurred.\n");
        /* Finalize options. */
        if (0 != rmixregistryd_options_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize options)))
        }
        return EXIT_FAILURE;
      }
      /* Wait for SIGTERM to shutdown. */
      sigwait(&sigmask, &signum);
      /* Unexport rmixregistry. */
      result = EXIT_SUCCESS;
      for (index = 0; index < count; index++) {
        /* Unexport rmixregistry. */
        if (0 != rmixregistry_unexport(&localrefs[index])) {
          RMIX_LOG((RMIX_WARN(unable to unexport # %u), index))
          fprintf(stdout, "ERROR: Unable to unexport # %u.\n", index);
          result = EXIT_FAILURE;
        }
      }
      /* Destroy rmixregistry. */
      if (0 != rmixregistry_destroy(&rmixregistry)) {
        RMIX_LOG((RMIX_WARN(unable to destroy rmixregistry)))
        fprintf(stdout, "ERROR: An internal error has occurred.\n");
        /* Finalize rmix. */
        if (0 != rmix_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
        }
        /* Deallocate local registry references. */
        free(localrefs);
        /* Finalize options. */
        if (0 != rmixregistryd_options_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize options)))
        }
        return EXIT_FAILURE;
      }
      /* Finalize rmix. */
      if (0 != rmix_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize rmix)))
        fprintf(stdout, "ERROR: An internal error has occurred.\n");
        /* Deallocate local registry references. */
        free(localrefs);
        /* Finalize options. */
        if (0 != rmixregistryd_options_fini()) {
          RMIX_LOG((RMIX_WARN(unable to finalize options)))
        }
        return EXIT_FAILURE;
      }
      /* Deallocate local registry references. */
      free(localrefs);
      /* Finalize options. */
      if (0 != rmixregistryd_options_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize options)))
        fprintf(stdout, "ERROR: An internal error has occurred.\n");
        return EXIT_FAILURE;
      }
      return result;
    }
    /* Check for parent process. */
    default: {
      /* Finalize options. */
      if (0 != rmixregistryd_options_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize options)))
      }
      return EXIT_SUCCESS;
    }
  }
}

/*
 * The signal handler.
 *
 * signum = The number of the received signal.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistry_handler"
void rmixregistry_handler (int signum) {
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
